"""
Serializers for ORA's BFF.

These are the response shapes that power the MFE implementation of the ORA UI.
"""
# pylint: disable=abstract-method

from rest_framework.serializers import (
    BooleanField,
    DateTimeField,
    IntegerField,
    Serializer,
    CharField,
    ListField,
    SerializerMethodField,
)


class CharListField(ListField):
    child = CharField()


class IsRequiredField(BooleanField):
    """
    Utility for checking if a field is "required" to reduce repeated code.
    """

    def to_representation(self, value):
        return value == "required"


class TextResponseConfigSerializer(Serializer):
    enabled = SerializerMethodField()
    required = IsRequiredField(source="text_response")
    editorType = CharField(source="text_response_editor")
    allowLatexPreview = BooleanField(source="allow_latex")

    def get_enabled(self, block):
        return block.text_response is not None


class FileResponseConfigSerializer(Serializer):
    enabled = SerializerMethodField()
    required = IsRequiredField(source="file_upload_response")
    fileUploadLimit = SerializerMethodField()
    allowedExtensions = CharListField(source="get_allowed_file_types_or_preset")
    blockedExtensions = CharListField(source="FILE_EXT_BLACK_LIST")
    fileTypeDescription = CharField(source="file_upload_type")

    def get_enabled(self, block):
        return block.file_upload_response is not None

    def get_fileUploadLimit(self, block):
        if not block.allow_multiple_files:
            return 1
        return block.MAX_FILES_COUNT


class TeamsConfigSerializer(Serializer):
    enabled = BooleanField(source="is_team_assignment")
    teamsetName = SerializerMethodField()

    def get_teamsetName(self, block):
        if block.teamset_config is not None:
            return block.teamset_config.name
        return None


class SubmissionConfigSerializer(Serializer):
    startDatetime = DateTimeField(source="submission_start")
    endDatetime = DateTimeField(source="submission_due")

    textResponseConfig = TextResponseConfigSerializer(source="*")
    fileResponseConfig = FileResponseConfigSerializer(source="*")

    teamsConfig = TeamsConfigSerializer(source="*")


class RubricFeedbackConfigSerializer(Serializer):
    description = CharField(source="rubric_feedback_prompt")  # is this this field?
    defaultText = CharField(source="rubric_feedback_default_text")


class RubricCriterionOptionSerializer(Serializer):
    name = CharField()
    label = CharField()
    points = IntegerField()
    description = CharField(source="explanation")


class RubricCriterionSerializer(Serializer):
    name = CharField(source="label")
    description = CharField(source="prompt")
    feedbackEnabled = SerializerMethodField()
    feedbackRequired = IsRequiredField(source="feedback")
    options = RubricCriterionOptionSerializer(many=True)

    @staticmethod
    def _feedback(criterion):
        # Feedback is disabled as a default
        return criterion.get("feedback", "disabled")

    def get_feedbackEnabled(self, criterion):
        # Feedback can be specified as optional or required
        return self._feedback(criterion) != "disabled"


class RubricConfigSerializer(Serializer):
    showDuringResponse = BooleanField(source="show_rubric_during_response")
    feedbackConfig = RubricFeedbackConfigSerializer(source="*")
    criteria = RubricCriterionSerializer(
        many=True, source="rubric_criteria_with_labels"
    )


class SelfSettingsSerializer(Serializer):
    required = BooleanField(default=True)

    startTime = DateTimeField(source="start")
    endTime = DateTimeField(source="due")


class PeerSettingsSerializer(Serializer):
    required = BooleanField(default=True)

    startTime = DateTimeField(source="start")
    endTime = DateTimeField(source="due")

    minNumberToGrade = IntegerField(source="must_grade")
    minNumberToBeGradedBy = IntegerField(source="must_be_graded_by")

    enableFlexibleGrading = BooleanField(
        source="enable_flexible_grading", required=False
    )


class AssessmentStepSettingsSerializer(Serializer):
    """
    Generic Assessments step, where we just need to know if the step is
    required given the ora.rubric_assessments source.
    """

    required = BooleanField(default=True)

    def _get_step(self, rubric_assessments, step_name):
        """Get the assessment step config for a given step_name"""
        for step in rubric_assessments:
            if step["name"] == step_name:
                return step
        return None

    def __init__(self, *args, **kwargs):
        self.step_name = kwargs.pop("step_name")
        super().__init__(*args, **kwargs)

    def to_representation(self, instance):
        assessment_step = self._get_step(instance, self.step_name)

        # Special handling for the peer step which includes extra fields
        if assessment_step and self.step_name == "peer-assessment":
            return PeerSettingsSerializer(assessment_step).data
        elif assessment_step and self.step_name == "self-assessment":
            return SelfSettingsSerializer(assessment_step).data

        # If we didn't find a step, it is not required
        if assessment_step is None:
            assessment_step = {"required": False}

        return super().to_representation(assessment_step)


class AssessmentStepsSettingsSerializer(Serializer):
    training = AssessmentStepSettingsSerializer(
        step_name="student-training", source="rubric_assessments"
    )
    peer = AssessmentStepSettingsSerializer(
        step_name="peer-assessment", source="rubric_assessments"
    )
    # Workaround to allow reserved keyword in serializer key
    vars()["self"] = AssessmentStepSettingsSerializer(
        step_name="self-assessment", source="rubric_assessments"
    )
    staff = AssessmentStepSettingsSerializer(
        step_name="staff-assessment", source="rubric_assessments"
    )


class AssessmentStepsSerializer(Serializer):
    order = SerializerMethodField()
    settings = AssessmentStepsSettingsSerializer(source="*")

    def get_order(self, block):
        return [step["name"] for step in block.rubric_assessments]


class LeaderboardConfigSerializer(Serializer):
    enabled = SerializerMethodField()
    numberOfEntries = IntegerField(source="leaderboard_show")

    def get_enabled(self, block):
        return block.leaderboard_show > 0


class OraBlockInfoSerializer(Serializer):
    """
    Main serializer for statically-defined ORA Block information
    """

    title = CharField()
    prompts = SerializerMethodField(source="*")
    baseAssetUrl = SerializerMethodField(source="*")

    submissionConfig = SubmissionConfigSerializer(source="*")
    assessmentSteps = AssessmentStepsSerializer(source="*")
    rubricConfig = RubricConfigSerializer(source="*")
    leaderboardConfig = LeaderboardConfigSerializer(source="*")

    def get_baseAssetUrl(self, block):
        # pylint: disable=protected-access
        return block.get_base_url_path_for_course_assets(
            block.course.id
        )

    def get_prompts(self, block):
        return [prompt["description"] for prompt in block.prompts]



""" Custom django template tags. """


from django import template
from django.template.defaultfilters import linebreaks, stringfilter
from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe

import bleach
from bleach import callbacks

register = template.Library()  # pylint: disable=invalid-name


@register.filter()
@stringfilter
def link_and_linebreak(text):
    """
    Converts URLs in text into clickable links with their target attribute set to `_blank`.
    It wraps givent tags into <p> tags and converts line breaks(\n) to <br> tags.
    Args:
        text: (str) Text having URLs to be converted
    Returns: (str) Text with URLs convert to links
    """
    if text:
        escaped_text = conditional_escape(text)
        return mark_safe(linebreaks(bleach.linkify(escaped_text, callbacks=[callbacks.target_blank])))
    return None



"""
Django admin models for openassessment
"""


import json

from django.contrib import admin
from django.urls import reverse_lazy
from django.utils.html import format_html, format_html_join

from openassessment.assessment.models import (
    Assessment, AssessmentFeedback, PeerWorkflow, PeerWorkflowItem, Rubric, SharedFileUpload, StaffWorkflow,
    TeamStaffWorkflow,
)
from openassessment.assessment.serializers import RubricSerializer


class RubricAdmin(admin.ModelAdmin):
    """
    Django admin model for Rubrics.
    """
    list_per_page = 20  # Loads of criteria summary are moderately expensive

    list_display = ('id', 'content_hash', 'criteria_summary')
    list_display_links = ('id', 'content_hash')
    search_fields = ('id', 'content_hash')
    readonly_fields = (
        'id', 'content_hash', 'structure_hash',
        'points_possible', 'criteria_summary', 'data'
    )

    def criteria_summary(self, rubric_obj):
        """Short description of criteria for presenting in a list."""
        rubric_data = RubricSerializer.serialized_from_cache(rubric_obj)
        return ", ".join(
            "{} - {}: {}".format(criterion["name"], criterion['label'], criterion["points_possible"])
            for criterion in rubric_data["criteria"]
        )

    def data(self, rubric_obj):
        """Full JSON string of rubric, indented and HTML formatted."""
        rubric_data = RubricSerializer.serialized_from_cache(rubric_obj)
        return format_html(
            "<pre>\n{}\n</pre>", json.dumps(rubric_data, sort_keys=True, indent=4))


class PeerWorkflowItemInline(admin.StackedInline):
    """
    Django admin model for PeerWorkflowItems.
    """
    model = PeerWorkflowItem
    fk_name = 'author'
    raw_id_fields = ('author', 'scorer', 'assessment')
    extra = 0


class PeerWorkflowAdmin(admin.ModelAdmin):
    """
    Django admin model for PeerWorkflows.
    """
    list_display = (
        'id', 'student_id', 'item_id', 'course_id', 'submission_uuid',
        'created_at', 'completed_at', 'grading_completed_at',
    )
    search_fields = (
        'id', 'student_id', 'item_id', 'course_id', 'submission_uuid',
    )
    inlines = (PeerWorkflowItemInline,)


class AssessmentAdmin(admin.ModelAdmin):
    """
    Django admin model for Assessments.
    """
    list_display = (
        'id', 'submission_uuid', 'score_type', 'scorer_id', 'scored_at',
        'rubric_link',
    )
    search_fields = (
        'id', 'submission_uuid', 'score_type', 'scorer_id', 'scored_at',
        'rubric__content_hash',
    )
    readonly_fields = (
        'submission_uuid', 'rubric_link', 'scored_at', 'scorer_id',
        'score_type', 'points_earned', 'points_possible', 'feedback',
        'parts_summary',
    )
    exclude = ('rubric', 'submission_uuid')

    def rubric_link(self, assessment_obj):
        """
        Returns the rubric link for this assessment.
        """
        url = reverse_lazy(
            'admin:assessment_rubric_change',
            args=[assessment_obj.rubric.id]
        )
        return format_html(
            '<a href="{}">{}</a>', url, assessment_obj.rubric.content_hash)
    rubric_link.admin_order_field = 'rubric__content_hash'
    rubric_link.short_description = 'Rubric'

    def parts_summary(self, assessment_obj):
        """
        Returns the parts summary of this assessment as HTML.
        """
        return format_html_join("<br/>", "{}/{} - {} - {}: {} - {} - {}", ((
            part.points_earned,
            part.points_possible,
            part.criterion.name,
            part.criterion.label,
            part.option.name if part.option else "None",
            part.option.label if part.option else "None",
            part.feedback
        ) for part in assessment_obj.parts.all()))


class AssessmentFeedbackAdmin(admin.ModelAdmin):
    """
    Django admin model for AssessmentFeedbacks.
    """
    list_display = ('id', 'submission_uuid',)
    search_fields = ('id', 'submission_uuid',)
    readonly_fields = (
        'submission_uuid', 'assessments_by', 'options', 'feedback_text'
    )
    exclude = ('assessments',)

    def assessments_by(self, assessment_feedback):
        """
        Gets all assessments for this feedback.
        """
        return format_html_join(", ", '<a href="{}">{}</a>', ((
            reverse_lazy('admin:assessment_assessment_change', args=[asmt.id]),
            asmt.scorer_id) for asmt in assessment_feedback.assessments.all()
        ))


class SharedFileUploadAdmin(admin.ModelAdmin):
    """
    Django admin model for SharedFileUploads.
    """
    model = SharedFileUpload
    list_display = (
        'id', 'team_id', 'item_id', 'owner_id', 'file_key', 'name',
        'description', 'size',
    )
    search_fields = (
        'id', 'course_id', 'item_id', 'team_id', 'file_key',
    )


class StaffWorkflowAdmin(admin.ModelAdmin):
    """
    Django admin model for StaffWorkflows
    """
    list_display = ('id', 'submission_uuid', 'course_id', 'item_id', 'grading_completed_at')


class TeamStaffWorkflowAdmin(admin.ModelAdmin):
    """
    Django admin model for TeamStaffWorkflows
    """
    list_display = ('id', 'team_submission_uuid', 'course_id', 'item_id', 'grading_completed_at')


admin.site.register(Rubric, RubricAdmin)
admin.site.register(PeerWorkflow, PeerWorkflowAdmin)
admin.site.register(Assessment, AssessmentAdmin)
admin.site.register(AssessmentFeedback, AssessmentFeedbackAdmin)
admin.site.register(SharedFileUpload, SharedFileUploadAdmin)
admin.site.register(StaffWorkflow, StaffWorkflowAdmin)
admin.site.register(TeamStaffWorkflow, TeamStaffWorkflowAdmin)



"""
Schema for validating and sanitizing data received from the JavaScript client.
"""

import dateutil
from pytz import utc

from voluptuous import (
    All,
    Any,
    In,
    Invalid,
    Optional,
    Range,
    Required,
    Schema,
)
from openassessment.xblock.utils.defaults import DATE_CONFIG_MANUAL, DATE_CONFIG_SUBSECTION, DATE_CONFIG_COURSE_END
from openassessment.xblock.utils.editor_config import AVAILABLE_EDITORS


def utf8_validator(value):
    """Validate and sanitize unicode strings.
    If we're given a bytestring, assume that the encoding is UTF-8

    Args:
        value: The value to validate

    Returns:
        unicode

    Raises:
        Invalid

    """
    try:
        if isinstance(value, bytes):
            return value.decode('utf-8')
        return str(value)
    except (ValueError, TypeError) as ex:
        raise Invalid(f"Could not load unicode from value \"{value}\"") from ex


def datetime_validator(value):
    """Validate and sanitize a datetime string in ISO format.

    Args:
        value: The value to validate

    Returns:
        unicode: ISO-formatted datetime string

    Raises:
        Invalid

    """
    try:
        # The dateutil parser defaults empty values to the current day,
        # which is NOT what we want.
        if value is None or value == '':
            raise Invalid(f"Datetime value cannot be \"{value}\"")

        # Parse the date and interpret it as UTC
        value = dateutil.parser.parse(value).replace(tzinfo=utc)
        return str(value.isoformat())
    except (ValueError, TypeError) as ex:
        raise Invalid(f"Could not parse datetime from value \"{value}\"") from ex


PROMPTS_TYPES = [
    'text',
    'html',
]


NECESSITY_OPTIONS = [
    'required',
    'optional',
    ''
]

# Build editor options from AVAILABLE_EDITORS
AVAILABLE_EDITOR_OPTIONS = set(AVAILABLE_EDITORS.keys())

VALID_ASSESSMENT_TYPES = [
    'peer-assessment',
    'self-assessment',
    'student-training',
    'staff-assessment',
]

VALID_UPLOAD_FILE_TYPES = [
    'image',
    'pdf-and-image',
    'custom'
]

DATE_CONFIG_OPTIONS = [
    DATE_CONFIG_MANUAL,
    DATE_CONFIG_SUBSECTION,
    DATE_CONFIG_COURSE_END,
]

# Schema definition for an update from the Studio JavaScript editor.
EDITOR_UPDATE_SCHEMA = Schema({
    Required('prompts'): [
        Schema({
            Required('description'): utf8_validator,
        })
    ],
    Required('prompts_type', default='text'): Any(All(utf8_validator, In(PROMPTS_TYPES)), None),
    Required('title'): utf8_validator,
    Required('feedback_prompt'): utf8_validator,
    Required('feedback_default_text'): utf8_validator,
    Required('submission_start'): Any(datetime_validator, None),
    Required('submission_due'): Any(datetime_validator, None),
    Required('date_config_type'): Any(All(utf8_validator, In(DATE_CONFIG_OPTIONS)), None),
    Required('text_response', default='required'): Any(All(utf8_validator, In(NECESSITY_OPTIONS)), None),
    Required('text_response_editor', default='text'): Any(All(utf8_validator, In(AVAILABLE_EDITOR_OPTIONS)), None),
    Required('file_upload_response', default=None): Any(All(utf8_validator, In(NECESSITY_OPTIONS)), None),
    'allow_file_upload': bool,  # Backwards compatibility.
    Required('file_upload_type', default=None): Any(All(utf8_validator, In(VALID_UPLOAD_FILE_TYPES)), None),
    'white_listed_file_types': utf8_validator,
    Required('allow_multiple_files'): bool,
    Required('allow_latex'): bool,
    Required('leaderboard_show'): int,
    Optional('teams_enabled'): bool,
    Optional('selected_teamset_id'): utf8_validator,
    Required('assessments'): [
        Schema({
            Required('name'): All(utf8_validator, In(VALID_ASSESSMENT_TYPES)),
            Required('start', default=None): Any(datetime_validator, None),
            Required('due', default=None): Any(datetime_validator, None),
            'required': bool,
            'must_grade': All(int, Range(min=0)),
            'must_be_graded_by': All(int, Range(min=0)),
            Required('enable_flexible_grading', default=False): bool,
            'examples': [
                Schema({
                    Required('answer'): [utf8_validator],
                    Required('options_selected'): [
                        Schema({
                            Required('criterion'): utf8_validator,
                            Required('option'): utf8_validator
                        })
                    ]
                })
            ],
            'examples_xml': utf8_validator,
        })
    ],
    Required('editor_assessments_order'): [
        All(utf8_validator, In(VALID_ASSESSMENT_TYPES))
    ],
    Required('feedbackprompt', default=""): utf8_validator,
    Required('criteria'): [
        Schema({
            Required('order_num'): All(int, Range(min=0)),
            Required('name'): utf8_validator,
            Required('label'): utf8_validator,
            Required('prompt'): utf8_validator,
            Required('feedback'): All(
                utf8_validator,
                In([
                    'disabled',
                    'optional',
                    'required',
                ])
            ),
            Required('options'): [
                Schema({
                    Required('order_num'): All(int, Range(min=0)),
                    Required('name'): utf8_validator,
                    Required('label'): utf8_validator,
                    Required('explanation'): utf8_validator,
                    Required('points'): All(int, Range(min=0)),
                })
            ]
        })
    ],
    Required('show_rubric_during_response', default=False): bool,
})



"""
External API for ORA Peer Assessment data
"""
import logging

from openassessment.assessment.errors import PeerAssessmentWorkflowError
from openassessment.assessment.api import peer as peer_api
from openassessment.xblock.utils.data_conversion import create_submission_dict
from openassessment.xblock.apis.step_data_api import StepDataAPI

logger = logging.getLogger(__name__)


class PeerAssessmentAPI(StepDataAPI):
    def __init__(self, block, continue_grading=False):
        super().__init__(block, "peer-assessment")
        self._continue_grading = continue_grading

    @property
    def assessment(self):
        return self.config_data.get_assessment_module("peer-assessment")

    @property
    def continue_grading(self):
        return self._continue_grading and self.workflow_data.is_peer_complete

    @property
    def file_upload_type(self):
        return self.config_data.file_upload_type

    @property
    def has_finished(self):
        finished, count = peer_api.has_finished_required_evaluating(
            self._block.submission_uuid, self.assessment["must_grade"]
        )
        return finished, count

    @property
    def is_cancelled(self):
        return self.workflow_data.is_cancelled

    @property
    def is_complete(self):
        finished = False
        if self.assessment:
            finished = self.has_finished[0]
        return self.workflow_data.is_done or finished

    @property
    def is_peer(self):
        return self.workflow_data.is_peer

    @property
    def is_skipped(self):
        return self.workflow_data.is_peer_skipped

    @property
    def student_item(self):
        return self.config_data.student_item_dict

    def format_submission_for_publish(self, peer_submission):
        student_item_dict = self.config_data.student_item_dict
        return {
            "requesting_student_id": student_item_dict["student_id"],
            "course_id": student_item_dict["course_id"],
            "item_id": student_item_dict["item_id"],
            "submission_returned_uuid": peer_submission["uuid"] if peer_submission else None,
        }

    def get_submission_dict(self, peer_sub):
        return create_submission_dict(peer_sub, self.config_data.prompts)

    def get_download_urls(self, peer_sub):
        return self._block.get_download_urls_from_submission(peer_sub)

    def get_peer_submission(self):
        peer_submission = False
        try:
            peer_submission = peer_api.get_submission_to_assess(
                self.workflow_data.submission_uuid, self.assessment["must_be_graded_by"]
            )
            self.config_data.publish_event(
                "openassessmentblock.get_peer_submission",
                self.format_submission_for_publish(peer_submission),
            )
        except PeerAssessmentWorkflowError as err:
            logger.exception(err)
        return peer_submission



"""
Generate CSV files for submission and assessment data, then upload to S3.
"""


import datetime
import os
import os.path
import shutil
import sys
import tarfile
import tempfile

from django.core.management.base import BaseCommand, CommandError

from openassessment.data import CsvWriter
from openassessment.fileupload.backends.s3 import _connect_to_s3


class Command(BaseCommand):
    """
    Create and upload CSV files for submission and assessment data.
    """

    help = 'Create and upload CSV files for submission and assessment data.'
    args = '<COURSE_ID> <S3_BUCKET_NAME>'

    OUTPUT_CSV_PATHS = {
        output_name: f"{output_name}.csv"
        for output_name in CsvWriter.MODELS
    }

    URL_EXPIRATION_HOURS = 24
    PROGRESS_INTERVAL = 10

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._history = []
        self._submission_counter = 0

    @property
    def history(self):
        """
        Return the upload history, which is useful for testing.

        Returns:
            list of dictionaries with keys 'url' and 'key'

        """
        return self._history

    def handle(self, *args, **options):
        """
        Execute the command.

        Args:
            course_id (unicode): The ID of the course to use.
            s3_bucket_name (unicode): The name of the S3 bucket to upload to.

        Raises:
            CommandError

        """
        if len(args) < 2:
            raise CommandError(f'Usage: upload_oa_data {self.args}')

        course_id, s3_bucket = args[0], args[1]
        if isinstance(course_id, bytes):
            course_id = course_id.decode('utf-8')
        if isinstance(s3_bucket, bytes):
            s3_bucket = s3_bucket.decode('utf-8')
        csv_dir = tempfile.mkdtemp()

        try:
            print(f"Generating CSV files for course '{course_id}'")
            self._dump_to_csv(course_id, csv_dir)
            print(f"Creating archive of CSV files in {csv_dir}")
            archive_path = self._create_archive(csv_dir)
            print(f"Uploading {archive_path} to {s3_bucket}/{course_id}")
            url = self._upload(course_id, archive_path, s3_bucket)
            print("== Upload successful ==")
            print(f"Download URL (expires in {self.URL_EXPIRATION_HOURS} hours):\n{url}")
        finally:
            # Assume that the archive was created in the directory,
            # so to clean up we just need to delete the directory.
            shutil.rmtree(csv_dir)

    def _dump_to_csv(self, course_id, csv_dir):
        """
        Create CSV files for submission/assessment data in a directory.

        Args:
            course_id (unicode): The ID of the course to dump data from.
            csv_dir (unicode): The absolute path to the directory in which to create CSV files.

        Returns:
            None
        """
        output_streams = {
            name: open(os.path.join(csv_dir, rel_path), 'w')  # pylint: disable=consider-using-with
            for name, rel_path in self.OUTPUT_CSV_PATHS.items()
        }
        csv_writer = CsvWriter(output_streams, self._progress_callback)
        csv_writer.write_to_csv(course_id)

    def _create_archive(self, dir_path):
        """
        Create an archive of a directory.

        Args:
            dir_path (unicode): The absolute path to the directory containing the CSV files.

        Returns:
            unicode: Absolute path to the archive.

        """
        tarball_name = "{}.tar.gz".format(
            datetime.datetime.utcnow().strftime("%Y-%m-%dT%H_%M")
        )
        tarball_path = os.path.join(dir_path, tarball_name)
        with tarfile.open(tarball_path, "w:gz") as tar:
            for rel_path in self.OUTPUT_CSV_PATHS.values():
                tar.add(os.path.join(dir_path, rel_path), arcname=rel_path)
        return tarball_path

    def _upload(self, course_id, file_path, s3_bucket):
        """
        Upload a file.

        Args:
            course_id (unicode): The ID of the course.
            file_path (unicode): Absolute path to the file to upload.
            s3_bucket (unicode): Name of the S3 bucket where the file will be uploaded.

        Returns:
            str: URL to access the uploaded archive.

        """
        conn = _connect_to_s3()

        key_name = os.path.join(course_id, os.path.split(file_path)[1])
        with open(file_path, "rb") as f:
            conn.put_object(
                Bucket=s3_bucket,
                Key=key_name,
                Body=f.read()
            )
        url = conn.generate_presigned_url(
            "get_object",
            Params={
                "Bucket": s3_bucket,
                "Key": key_name,
            },
            ExpiresIn=self.URL_EXPIRATION_HOURS * 3600
        )

        # Store the key and url in the history
        self._history.append({'key': key_name, 'url': url})

        return url

    def _progress_callback(self):
        """
        Indicate progress to the user as submissions are processed.
        """
        self._submission_counter += 1
        if self._submission_counter > 0 and self._submission_counter % self.PROGRESS_INTERVAL == 0:
            sys.stdout.write('.')
            sys.stdout.flush()



""" S3 bucket file upload backend. """


import logging

from django.conf import settings

import botocore
import boto3

from ..exceptions import FileUploadInternalError
from .base import BaseBackend

log = logging.getLogger(
    "openassessment.fileupload.api"
)  # pylint: disable=invalid-name


class Backend(BaseBackend):
    """ S3 Bucked File Upload Backend. """

    def get_upload_url(self, key, content_type):
        bucket_name, key_name = self._retrieve_parameters(key)
        try:
            conn = _connect_to_s3()
            return conn.generate_presigned_url(
                "put_object",
                Params={
                    "Bucket": bucket_name,
                    "Key": key_name,
                    "ContentType": content_type,
                },
                ExpiresIn=self.UPLOAD_URL_TIMEOUT,
            )
        except Exception as ex:
            log.exception(
                "An internal exception occurred while generating an upload URL."
            )
            raise FileUploadInternalError(ex) from ex

    def get_download_url(self, key):
        bucket_name, key_name = self._retrieve_parameters(key)
        try:
            conn = _connect_to_s3()
            if not object_exists(conn, bucket_name, key_name):
                return ""
            return conn.generate_presigned_url(
                "get_object",
                Params={"Bucket": bucket_name, "Key": key_name},
                ExpiresIn=self.DOWNLOAD_URL_TIMEOUT,
            )
        except Exception as ex:
            log.exception(
                "An internal exception occurred while generating a download URL."
            )
            raise FileUploadInternalError(ex) from ex

    def remove_file(self, key):
        bucket_name, key_name = self._retrieve_parameters(key)
        conn = _connect_to_s3()
        if object_exists(conn, bucket_name, key_name):
            conn.delete_object(Bucket=bucket_name, Key=key_name)
            return True
        return False


def _connect_to_s3():
    """Connect to s3

    Creates a connection to s3 for file URLs.
    """
    # Try to get the AWS credentials from settings if they are available
    # If not, these will default to `None`, and boto3 will try to use
    # environment vars or configuration files instead.
    aws_access_key_id = getattr(settings, "AWS_ACCESS_KEY_ID", None)
    aws_secret_access_key = getattr(settings, "AWS_SECRET_ACCESS_KEY", None)
    endpoint_url = getattr(settings, "AWS_S3_ENDPOINT_URL", None)
    signature_version = getattr(settings, "AWS_S3_SIGNATURE_VERSION", None)
    region_name = getattr(settings, "AWS_S3_REGION_NAME", None)

    return boto3.client(
        "s3",
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        endpoint_url=endpoint_url,
        config=botocore.client.Config(
            signature_version=signature_version,
            region_name=region_name,
        )
    )


def object_exists(conn, bucket_name, key_name):
    """
    Check if a key exists in the given S3 bucket.
    """
    try:
        conn.head_object(Bucket=bucket_name, Key=key_name)
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "404":
            return False
        raise e
    return True



"""An XBlock where students can read a question and compose their response"""

import copy
import datetime as dt
from functools import cached_property
import json
import logging
import re

import pkg_resources
import pytz

from django.conf import settings
from django.template.loader import get_template

from bleach.sanitizer import Cleaner
from lazy import lazy
from webob import Response
from xblock.core import XBlock
from xblock.exceptions import NoSuchServiceError
from xblock.fields import Boolean, Integer, List, Scope, String

from openassessment.staffgrader.staff_grader_mixin import StaffGraderMixin
from openassessment.workflow.errors import AssessmentWorkflowError
from openassessment.xblock.apis.submissions.submissions_api import SubmissionAPI
from openassessment.xblock.course_items_listing_mixin import CourseItemsListingMixin
from openassessment.xblock.utils.data_conversion import (
    create_prompts_list,
    create_rubric_dict,
    update_assessments_format,
)
from openassessment.xblock.utils.defaults import *  # pylint: disable=wildcard-import, unused-wildcard-import
from openassessment.xblock.files_mixin import FilesMixin
from openassessment.xblock.grade_mixin import GradeMixin
from openassessment.xblock.leaderboard_mixin import LeaderboardMixin
from openassessment.xblock.lms_mixin import LmsCompatibilityMixin
from openassessment.xblock.message_mixin import MessageMixin
from openassessment.xblock.mobile import togglable_mobile_support
from openassessment.xblock.utils.resolve_dates import (
    DateValidationError,
    DISTANT_FUTURE,
    DISTANT_PAST,
    parse_date_value,
    resolve_dates,
)
from openassessment.xblock.rubric_reuse_mixin import RubricReuseMixin
from openassessment.xblock.staff_area_mixin import StaffAreaMixin
from openassessment.xblock.studio_mixin import StudioMixin
from openassessment.xblock.team_mixin import TeamMixin
from openassessment.xblock.ui_mixins.legacy.handlers_mixin import LegacyHandlersMixin
from openassessment.xblock.ui_mixins.legacy.views_mixin import LegacyViewsMixin
from openassessment.xblock.ui_mixins.mfe.mixin import MfeMixin
from openassessment.xblock.utils.validation import validator
from openassessment.xblock.config_mixin import ConfigMixin
from openassessment.xblock.workflow_mixin import WorkflowMixin
from openassessment.xblock.team_workflow_mixin import TeamWorkflowMixin
from openassessment.xblock.openassesment_template_mixin import OpenAssessmentTemplatesMixin
from openassessment.xblock.utils.xml import parse_from_xml, serialize_content_to_xml


from openassessment.xblock.apis.ora_config_api import ORAConfigAPI
from openassessment.xblock.apis.workflow_api import WorkflowAPI
from openassessment.xblock.apis.assessments.peer_assessment_api import PeerAssessmentAPI
from openassessment.xblock.apis.assessments.self_assessment_api import SelfAssessmentAPI
from openassessment.xblock.apis.assessments.staff_assessment_api import StaffAssessmentAPI
from openassessment.xblock.apis.assessments.student_training_api import StudentTrainingAPI
from openassessment.xblock.apis.ora_data_accessor import ORADataAccessor

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def load(path):
    """Handy helper for getting resources from our kit."""
    data = pkg_resources.resource_string(__name__, path)
    return data.decode("utf8")


@XBlock.needs("i18n")
@XBlock.needs("user")
@XBlock.needs("user_state")
@XBlock.needs("teams")
@XBlock.needs("teams_configuration")
class OpenAssessmentBlock(
    MessageMixin,
    StudioMixin,
    FilesMixin,
    GradeMixin,
    LeaderboardMixin,
    StaffAreaMixin,
    WorkflowMixin,
    TeamWorkflowMixin,
    LmsCompatibilityMixin,
    CourseItemsListingMixin,
    ConfigMixin,
    TeamMixin,
    OpenAssessmentTemplatesMixin,
    RubricReuseMixin,
    StaffGraderMixin,
    MfeMixin,
    LegacyViewsMixin,
    LegacyHandlersMixin,
    XBlock,
):
    """Displays a prompt and provides an area where students can compose a response."""

    VALID_ASSESSMENT_TYPES = [
        "student-training",
        "peer-assessment",
        "self-assessment",
        "staff-assessment",
    ]

    VALID_ASSESSMENT_TYPES_FOR_TEAMS = [  # pylint: disable=invalid-name
        'staff-assessment',
    ]

    public_dir = 'static'

    allow_latex = Boolean(
        default=False,
        scope=Scope.settings,
        help="Latex rendering allowed with submission."
    )

    allow_file_upload = Boolean(
        default=False,
        scope=Scope.content,
        help="Do not use. For backwards compatibility only."
    )

    allow_multiple_files = Boolean(
        default=True,
        scope=Scope.settings,
        help="Allow multiple files uploaded with submission (if file upload enabled)."
    )

    date_config_type = String(
        default=DATE_CONFIG_MANUAL,
        scope=Scope.settings,
        help="The type of date configuration. Possible values are 'manual', 'subsection', and 'course_end'."
    )

    file_upload_response_raw = String(
        help="Specify whether learners are able to upload files as a part of their response.",
        default=None,
        scope=Scope.settings
    )

    file_upload_type_raw = String(
        default=None,
        scope=Scope.content,
        help="File upload to be included with submission (can be 'image', 'pdf-and-image', or 'custom')."
    )

    has_saved = Boolean(
        default=False,
        scope=Scope.user_state,
        help="Indicates whether the user has saved a response."
    )

    leaderboard_show = Integer(
        default=0,
        scope=Scope.content,
        help="The number of leaderboard results to display (0 if none)"
    )

    no_peers = Boolean(
        default=False,
        scope=Scope.user_state,
        help="Indicates whether or not there are peers to grade."
    )

    prompt = String(
        default=DEFAULT_PROMPT,
        scope=Scope.content,
        help="The prompts to display to a student."
    )

    prompts_type = String(
        default='text',
        scope=Scope.content,
        help="The type of prompt. html or text"
    )

    rubric_criteria = List(
        default=DEFAULT_RUBRIC_CRITERIA,
        scope=Scope.content,
        help="The different parts of grading for students giving feedback."
    )

    rubric_feedback_prompt = String(
        default=DEFAULT_RUBRIC_FEEDBACK_PROMPT,
        scope=Scope.content,
        help="The rubric feedback prompt displayed to the student"
    )

    rubric_feedback_default_text = String(
        default=DEFAULT_RUBRIC_FEEDBACK_TEXT,
        scope=Scope.content,
        help="The default rubric feedback text displayed to the student"
    )

    rubric_assessments = List(
        default=DEFAULT_ASSESSMENT_MODULES,
        scope=Scope.content,
        help="The requested set of assessments and the order in which to apply them."
    )

    saved_files_descriptions = String(
        default="",
        scope=Scope.user_state,
        help="Saved descriptions for each uploaded file."
    )

    saved_files_names = String(
        default="",
        scope=Scope.user_state,
        help="Saved original names for each uploaded file."
    )

    saved_files_sizes = String(
        default="",
        scope=Scope.user_state,
        help="Filesize of each uploaded file in bytes."
    )

    saved_response = String(
        default="",
        scope=Scope.user_state,
        help="Saved response submission for the current user."
    )

    selected_teamset_id = String(
        default="",
        scope=Scope.settings,
        help="The id of the selected teamset.",
    )

    show_rubric_during_response = Boolean(
        default=False,
        scope=Scope.settings,
        help="Should the rubric be visible to learners in the response section?"
    )

    submission_due = String(
        default=DEFAULT_DUE, scope=Scope.settings,
        help="ISO-8601 formatted string representing the submission due date."
    )

    submission_start = String(
        default=DEFAULT_START, scope=Scope.settings,
        help="ISO-8601 formatted string representing the submission start date."
    )

    submission_uuid = String(
        default=None,
        scope=Scope.user_state,
        help="The student's submission that others will be assessing."
    )

    teams_enabled = Boolean(
        default=False,
        scope=Scope.settings,
        help="Whether team submissions are enabled for this case study.",
    )

    text_response_raw = String(
        help="Specify whether learners must include a text based response to this problem's prompt.",
        default="required",
        scope=Scope.settings
    )

    text_response_editor = String(
        help="Select which editor learners will use to include a text based response to this problem's prompt.",
        scope=Scope.settings,
        default='text'
    )

    title = String(
        default="Open Response Assessment",
        scope=Scope.content,
        help="A title to display to a student (plain text)."
    )

    white_listed_file_types = List(
        default=[],
        scope=Scope.content,
        help="Custom list of file types allowed with submission."
    )

    @property
    def config_data(self):
        return ORAConfigAPI(self)

    @property
    def workflow_data(self):
        return WorkflowAPI(self)

    @property
    def submission_data(self):
        return SubmissionAPI(self)

    def peer_assessment_data(self, continue_grading=False):
        return PeerAssessmentAPI(self, continue_grading)

    @property
    def self_data(self):
        return SelfAssessmentAPI(self)

    @property
    def staff_data(self):
        return StaffAssessmentAPI(self)

    @property
    def student_training_data(self):
        return StudentTrainingAPI(self)

    @property
    def api_data(self):
        return ORADataAccessor(self)

    @property
    def course_id(self):
        return str(self.xmodule_runtime.course_id)  # pylint: disable=no-member

    @cached_property
    def course(self):
        if not hasattr(self.runtime, "modulestore"):
            return None
        return self.runtime.modulestore.get_course(self.scope_ids.usage_id.context_key)

    @property
    def text_response(self):
        """
        Backward compatibility for existing blocks that were created without text_response
        or file_upload_response fields. These blocks will be treated as required text.
        """
        if not self.file_upload_response_raw and not self.text_response_raw:
            return 'required'
        return self.text_response_raw

    @text_response.setter
    def text_response(self, value):
        """
        Setter for text_response_raw
        """
        self.text_response_raw = value if value else None

    @property
    def file_upload_response(self):
        """
        Backward compatibility for existing block before that were created without
        'text_response' and 'file_upload_response_raw' fields.
        """
        if not self.file_upload_response_raw and (self.file_upload_type_raw is not None or self.allow_file_upload):
            return 'optional'
        return self.file_upload_response_raw

    @file_upload_response.setter
    def file_upload_response(self, value):
        """
        Setter for file_upload_response_raw
        """
        self.file_upload_response_raw = value if value else None

    @property
    def file_upload_type(self):
        """
        Backward compatibility for existing block before the change from allow_file_upload to file_upload_type_raw.

        This property will use new file_upload_type_raw field when available, otherwise will fall back to
        allow_file_upload field for old blocks.
        """
        if self.file_upload_type_raw is not None:
            return self.file_upload_type_raw
        if self.allow_file_upload:
            return 'image'
        return None

    @file_upload_type.setter
    def file_upload_type(self, value):
        """
        Setter for file_upload_type_raw
        """
        self.file_upload_type_raw = value

    @property
    def white_listed_file_types_string(self):
        """
        Join the white listed file types into comma delimited string
        """
        if self.white_listed_file_types:
            return ','.join(self.white_listed_file_types)
        return ''

    @white_listed_file_types_string.setter
    def white_listed_file_types_string(self, value):
        """
        Convert comma delimited white list string into list with some clean up
        """
        self.white_listed_file_types = [file_type.strip().strip('.').lower()
                                        for file_type in value.split(',')] if value else None

    def get_anonymous_user_id(self, username, course_id):
        """
        Get the anonymous user id from Xblock user service.

        Args:
            username(str): user's name entered by staff to get info.
            course_id(str): course id.

        Returns:
            A unique id for (user, course) pair
        """
        return self.runtime.service(self, 'user').get_anonymous_user_id(username, course_id)

    def is_user_state_service_available(self):
        """
        Check if the user state service is present in runtime.
        """
        try:
            self.runtime.service(self, 'user_state')
            return True
        except NoSuchServiceError:
            return False

    def get_user_state(self, username):
        """
        Get the student module state for the given username for current ORA block.

        Arguments:
            username(str): username against which the state is required in the current block.

        Returns:
            user state, if found, else empty dict
        """
        if self.is_user_state_service_available():
            user_state_service = self.runtime.service(self, 'user_state')
            return user_state_service.get_state_as_dict(username, str(self.location))  # pylint: disable=no-member
        return {}

    def should_use_user_state(self, upload_urls):
        """
        Return a boolean if the user state is used for additional data checks.

        User state is utilized when all of the following are true:
        1. user state service is available(which is only part of courseware)
        2. The waffle flag/switch is enabled
        3. the file upload is required or optional
        4. the file data from submission is missing information
        """
        return not any(upload_urls) \
            and self.is_user_state_service_available() \
            and self.user_state_upload_data_enabled \
            and self.file_upload_response

    def should_get_all_files_urls(self, upload_urls):
        """
        Returns a boolean to decide if all the file submitted by a learner in a block should be obtained.

        Following conditions should be true for boolean to be true:
        1. The waffle flag/switch is enabled
        2. the file upload is required or optional
        3. the file data from submission is missing information

        Arguments:
            upload_urls(list): A list of (file url, description, name) tuple, if info present, else empty list
        """
        return not any(upload_urls) \
            and self.is_fetch_all_urls_waffle_enabled \
            and self.file_upload_response

    def get_student_item_dict_from_username_or_email(self, username_or_email):
        """
        Get the item dict for a given username or email in the parent course of block.
        """
        anonymous_user_id = self.get_anonymous_user_id(username_or_email, self.course_id)
        return self.get_student_item_dict(anonymous_user_id=anonymous_user_id)

    def get_anonymous_user_id_from_xmodule_runtime(self):
        if hasattr(self, "xmodule_runtime"):
            return self.xmodule_runtime.anonymous_student_id  # pylint:disable=E1101
        return None

    def get_student_item_dict(self, anonymous_user_id=None):
        """Create a student_item_dict from our surrounding context.

        See also: submissions.api for details.

        Args:
            anonymous_user_id(str): A unique anonymous_user_id for (user, course) pair.
        Returns:
            (dict): The student item associated with this XBlock instance. This
                includes the student id, item id, and course id.
        """

        item_id = str(self.scope_ids.usage_id)

        # This is not the real way course_ids should work, but this is a
        # temporary expediency for LMS integration
        if hasattr(self, "xmodule_runtime"):
            course_id = self.course_id
            if anonymous_user_id:
                student_id = anonymous_user_id
            else:
                student_id = self.xmodule_runtime.anonymous_student_id  # pylint:disable=E1101
        else:
            course_id = "edX/Enchantment_101/April_1"
            if self.scope_ids.user_id is None:
                student_id = None
            else:
                student_id = str(self.scope_ids.user_id)

        student_item_dict = {
            "student_id": student_id,
            "item_id": item_id,
            "course_id": course_id,
            "item_type": 'openassessment'
        }
        return student_item_dict

    @togglable_mobile_support
    def student_view(self, context=None):  # pylint: disable=unused-argument
        """The main view of OpenAssessmentBlock, displayed when viewing courses.

        The main view which displays the general layout for Open Ended
        Assessment Questions. The contents of the XBlock are determined
        dynamically based on the assessment workflow configured by the author.

        Args:
            context: Not used for this view.

        Returns:
            (Fragment): The HTML Fragment for this XBlock, which determines the
            general frame of the Open Ended Assessment Question.
        """
        # On page load, update the workflow status.
        # We need to do this here because peers may have graded us, in which
        # case we may have a score available.

        try:
            self.update_workflow_status()
        except AssessmentWorkflowError:
            # Log the exception, but continue loading the page
            logger.exception('An error occurred while updating the workflow on page load.')

        ui_models = self._create_ui_models()
        # All data we intend to pass to the front end.
        context_dict = {
            "title": self.title,
            "prompts": self.prompts,
            "prompts_type": self.prompts_type,
            "rubric_assessments": ui_models,
            "show_staff_area": self.is_course_staff and not self.in_studio_preview,
        }
        template = get_template("openassessmentblock/oa_base.html")
        return self._create_fragment(template, context_dict, initialize_js_func='OpenAssessmentBlock')

    def ora_blocks_listing_view(self, context=None):
        """This view is used in the Open Response Assessment tab in the LMS Instructor Dashboard
        to display all available course ORA blocks.

        Args:
            context: contains two items:
                "ora_items" - all course items with names and parents, example:
                    [{"parent_name": "Vertical name",
                      "name": "ORA Display Name",
                      "url_grade_available_responses": "/grade_available_responses_view",
                      "url_waiting_step_details": "/waiting_step_details_view",
                      "staff_assessment": false,
                      "parent_id": "vertical_block_id",
                      "url_base": "/student_view",
                      "id": "openassessment_block_id"
                     }, ...]
                "ora_item_view_enabled" - enabled LMS API endpoint to serve XBlock view or not

        Returns:
            (Fragment): The HTML Fragment for this XBlock.
        """
        ora_items = context.get('ora_items', []) if context else []
        ora_item_view_enabled = context.get('ora_item_view_enabled', False) if context else False
        context_dict = {
            "ora_items": json.dumps(ora_items),
            "ora_item_view_enabled": ora_item_view_enabled
        }

        template = get_template('openassessmentblock/instructor_dashboard/oa_listing.html')

        min_postfix = '.min' if settings.DEBUG else ''

        return self._create_fragment(
            template,
            context_dict,
            initialize_js_func='CourseOpenResponsesListingBlock',
            additional_css=["static/css/lib/backgrid/backgrid%s.css" % min_postfix],
            additional_js=["static/js/lib/backgrid/backgrid%s.js" % min_postfix],
            additional_js_context={
                "ENHANCED_STAFF_GRADER": self.is_enhanced_staff_grader_enabled,
                "ORA_GRADING_MICROFRONTEND_URL": getattr(settings, 'ORA_GRADING_MICROFRONTEND_URL', '')
            }
        )

    def grade_available_responses_view(self, context=None):  # pylint: disable=unused-argument
        """Grade Available Responses view.

        Auxiliary view which displays the staff grading area
        (used in the Open Response Assessment tab in the Instructor Dashboard of LMS)

        Args:
            context: Not used for this view.

        Returns:
            (Fragment): The HTML Fragment for this XBlock.
        """
        student_item = self.get_student_item_dict()
        staff_assessment_required = "staff-assessment" in self.assessment_steps

        context_dict = {
            "title": self.title,
            'staff_assessment_required': staff_assessment_required,
        }

        if staff_assessment_required:
            context_dict.update(
                self.get_staff_assessment_statistics_context(student_item["course_id"], student_item["item_id"])
            )

        template = get_template('openassessmentblock/instructor_dashboard/oa_grade_available_responses.html')

        return self._create_fragment(template, context_dict, initialize_js_func='StaffAssessmentBlock')

    def waiting_step_details_view(self, context=None):  # pylint: disable=unused-argument
        """
        Waiting Step Details view.

        Auxiliary view which displays a list of students "stuck" in the
        peer grading step waiting for a grade (used in the Open Response Assessment
        tab in the Instructor Dashboard of LMS).

        Args:
            context: Not used for this view.

        Returns:
            (Fragment): The HTML Fragment for this XBlock.
        """
        peer_assessment_required = "peer-assessment" in self.assessment_steps

        context_dict = {
            "title": self.title,
            "peer_assessment_required": peer_assessment_required,
        }

        if peer_assessment_required:
            context_dict['waiting_step_data_url'] = self.runtime.handler_url(
                self, "waiting_step_data",
            )

        template = get_template('openassessmentblock/instructor_dashboard/oa_waiting_step_details.html')

        return self._create_fragment(
            template,
            context_dict,
            initialize_js_func='WaitingStepDetailsBlock',
            additional_js_context=context_dict,
        )

    @property
    def is_admin(self):
        """
        Check whether the user has global staff permissions.

        Returns:
            bool
        """
        if hasattr(self, 'xmodule_runtime'):
            return getattr(self.xmodule_runtime, 'user_is_admin', False)  # pylint: disable=no-member
        return False

    @property
    def is_course_staff(self):
        """
        Check whether the user has course staff permissions for this XBlock.

        Returns:
            bool
        """
        if hasattr(self, 'xmodule_runtime'):
            return getattr(self.xmodule_runtime, 'user_is_staff', False)  # pylint: disable=no-member
        return False

    @property
    def is_beta_tester(self):
        """
        Check whether the user is a beta tester.

        Returns:
            bool
        """
        if hasattr(self, 'xmodule_runtime'):
            return getattr(self.xmodule_runtime, 'user_is_beta_tester', False)  # pylint: disable=no-member
        return False

    @property
    def in_studio_preview(self):
        """
        Check whether we are in Studio preview mode.

        Returns:
            bool

        """
        # When we're running in Studio Preview mode, the XBlock won't provide us with a user ID.
        # (Note that `self.xmodule_runtime` will still provide an anonymous
        # student ID, so we can't rely on that)
        return self.scope_ids.user_id is None

    @property
    def has_real_user(self):
        """
        Checks whether the runtime is tied to a real user

        Returns:
            bool
        """
        if hasattr(self, 'xmodule_runtime'):
            return self.xmodule_runtime.get_real_user is not None  # pylint: disable=no-member
        return False

    @staticmethod
    def workbench_scenarios():
        """A canned scenario for display in the workbench.

        These scenarios are only intended to be used for Workbench XBlock
        Development.

        """
        return [
            (
                "OpenAssessmentBlock File Upload: Images",
                load('static/xml/file_upload_image_only.xml')
            ),
            (
                "OpenAssessmentBlock File Upload: PDF and Images",
                load('static/xml/file_upload_pdf_and_image.xml')
            ),
            (
                "OpenAssessmentBlock File Upload: Custom File Types",
                load('static/xml/file_upload_custom.xml')
            ),
            (
                "OpenAssessmentBlock File Upload: allow_file_upload compatibility",
                load('static/xml/file_upload_compat.xml')
            ),
            (
                "OpenAssessmentBlock Unicode",
                load('static/xml/unicode.xml')
            ),
            (
                "OpenAssessmentBlock Poverty Rubric",
                load('static/xml/poverty_rubric_example.xml')
            ),
            (
                "OpenAssessmentBlock Leaderboard",
                load('static/xml/leaderboard.xml')
            ),
            (
                "OpenAssessmentBlock Leaderboard with Custom File Type",
                load('static/xml/leaderboard_custom.xml')
            ),
            (
                "OpenAssessmentBlock (Peer Only) Rubric",
                load('static/xml/poverty_peer_only_example.xml')
            ),
            (
                "OpenAssessmentBlock (Self Only) Rubric",
                load('static/xml/poverty_self_only_example.xml')
            ),
            (
                "OpenAssessmentBlock Censorship Rubric",
                load('static/xml/censorship_rubric_example.xml')
            ),
            (
                "OpenAssessmentBlock Promptless Rubric",
                load('static/xml/promptless_rubric_example.xml')
            ),
        ]

    @classmethod
    def parse_xml(cls, node, runtime, keys, id_generator):
        """Instantiate XBlock object from runtime XML definition.

        Inherited by XBlock core.

        """
        config = parse_from_xml(node)
        block = runtime.construct_xblock_from_class(cls, keys)

        xblock_validator = validator(block, block._, strict_post_release=False)
        xblock_validator(
            create_rubric_dict(config['prompts'], config['rubric_criteria']),
            config['rubric_assessments'],
            submission_start=config['submission_start'],
            submission_due=config['submission_due'],
            leaderboard_show=config['leaderboard_show']
        )

        block.allow_file_upload = config['allow_file_upload']
        block.allow_latex = config['allow_latex']
        block.allow_multiple_files = config['allow_multiple_files']
        block.file_upload_response = config['file_upload_response']
        block.file_upload_type = config['file_upload_type']
        block.group_access = config['group_access']
        block.leaderboard_show = config['leaderboard_show']
        block.prompts = config['prompts']
        block.prompts_type = config['prompts_type']
        block.rubric_criteria = config['rubric_criteria']
        block.rubric_feedback_prompt = config['rubric_feedback_prompt']
        block.rubric_feedback_default_text = config['rubric_feedback_default_text']
        block.rubric_assessments = config['rubric_assessments']
        block.selected_teamset_id = config['selected_teamset_id']
        block.show_rubric_during_response = config['show_rubric_during_response']
        block.submission_start = config['submission_start']
        block.submission_due = config['submission_due']
        block.teams_enabled = config['teams_enabled']
        block.text_response = config['text_response']
        block.text_response_editor = config['text_response_editor']
        block.title = config['title']
        block.white_listed_file_types_string = config['white_listed_file_types']
        return block

    @property
    def _(self):
        i18nService = self.runtime.service(self, 'i18n')  # pylint: disable=invalid-name
        return i18nService.ugettext

    @property
    def prompts(self):
        """
        Return the prompts.

        Initially a block had a single prompt which was saved as a simple
        string in the prompt field. Now prompts are saved as a serialized
        list of dicts in the same field. If prompt field contains valid json,
        parse and return it. Otherwise, assume it is a simple string prompt
        and return it in a list of dict.

        Returns:
            list of dict
        """
        return create_prompts_list(self.prompt)

    @prompts.setter
    def prompts(self, value):
        """
        Serialize the prompts and save to prompt field.

        Args:
            value (list of dict): The prompts to set.
        """

        if value is None:
            self.prompt = None
        elif len(value) == 1:
            # For backwards compatibility. To be removed after all code
            # is migrated to use prompts property instead of prompt field.
            self.prompt = value[0]['description']
        else:
            self.prompt = json.dumps(value)

    @property
    def valid_assessments(self):
        """
        Return a list of assessment dictionaries that we recognize.
        This allows us to gracefully handle situations in which unrecognized
        assessment types are stored in the XBlock field (e.g. because
        we roll back code after releasing a feature).

        Returns:
            list

        """
        assessment_types = self.VALID_ASSESSMENT_TYPES
        if self.teams_enabled:
            assessment_types = self.VALID_ASSESSMENT_TYPES_FOR_TEAMS

        _valid_assessments = [
            asmnt for asmnt in self.rubric_assessments
            if asmnt.get('name') in assessment_types
        ]
        return update_assessments_format(copy.deepcopy(_valid_assessments))

    @property
    def assessment_steps(self):
        """
        Return a list of assessment steps by name.
        Also filter out assessments that have required set to false and do not
        contain a staff grade override.

        Returns:
            list

        """
        assessment_steps = []
        for assessment in self.valid_assessments:
            if assessment['name'] == 'staff-assessment' and assessment["required"] is False:
                if not StaffAssessmentAPI.staff_assessment_exists(self.submission_uuid):
                    continue
            assessment_steps.append(assessment['name'])
        return assessment_steps

    @lazy
    def rubric_criteria_with_labels(self):
        """
        Backwards compatibility: We used to treat "name" as both a user-facing label
        and a unique identifier for criteria and options.
        Now we treat "name" as a unique identifier, and we've added an additional "label"
        field that we display to the user.
        If criteria/options in the problem definition do NOT have a "label" field
        (because they were created before this change),
        we create a new label that has the same value as "name".

        The result of this call is cached, so it should NOT be used in a runtime
        that can modify the XBlock settings (in the LMS, settings are read-only).

        Returns:
            list of criteria dictionaries

        """
        criteria = copy.deepcopy(self.rubric_criteria)
        for criterion in criteria:
            if 'label' not in criterion:
                criterion['label'] = criterion['name']
            for option in criterion['options']:
                if 'label' not in option:
                    option['label'] = option['name']
        return criteria

    def render_assessment(self, path, context_dict=None):
        """Render an Assessment Module's HTML

        Given the name of an assessment module, find it in the list of
        configured modules, and ask for its rendered HTML.

        Args:
            path (str): The path to the template used to render this HTML
                section.
            context_dict (dict): A dictionary of context variables used to
                populate this HTML section.

        Returns:
            (Response): A Response Object with the generated HTML fragment. This
                is intended for AJAX calls to load dynamically into a larger
                document.
        """
        if not context_dict:
            context_dict = {}

        context_dict['text_response_editor'] = self.text_response_editor

        template = get_template(path)
        return Response(template.render(context_dict), content_type='application/html', charset='UTF-8')

    def add_xml_to_node(self, node):
        """
        Serialize the XBlock to XML for exporting.
        """
        serialize_content_to_xml(self, node)

    def render_error(self, error_msg):
        """
        Render an error message.

        Args:
            error_msg (unicode): The error message to display.

        Returns:
            Response: A response object with an HTML body.
        """
        context = {'error_msg': error_msg}
        template = get_template('openassessmentblock/oa_error.html')
        return Response(template.render(context), content_type='application/html', charset='UTF-8')

    def is_closed(self, step=None, course_staff=None):
        """
        Checks if the question is closed.

        Determines if the start date is in the future or the end date has
            passed.  Optionally limited to a particular step in the workflow.

        Start/due dates do NOT apply to course staff, since course staff may need to get to
        the peer grading step AFTER the submission deadline has passed.
        This may not be necessary when we implement a grading interface specifically for course staff.

        Keyword Arguments:
            step (str): The step in the workflow to check.  Options are:
                None: check whether the problem as a whole is open.
                "submission": check whether the submission section is open.
                "peer-assessment": check whether the peer-assessment section is open.
                "self-assessment": check whether the self-assessment section is open.

            course_staff (bool): Whether to treat the user as course staff (disable start/due dates).
                If not specified, default to the current user's status.

        Returns:
            tuple of the form (is_closed, reason, start_date, due_date), where
                is_closed (bool): indicates whether the step is closed.
                reason (str or None): specifies the reason the step is closed ("start" or "due")
                start_date (datetime): is the start date of the step/problem.
                due_date (datetime): is the due date of the step/problem.

        Examples:
            >>> is_closed()
            False, None, datetime.datetime(2014, 3, 27, 22, 7, 38, 788861),
            datetime.datetime(2015, 3, 27, 22, 7, 38, 788861)
            >>> is_closed(step="submission")
            True, "due", datetime.datetime(2014, 3, 27, 22, 7, 38, 788861),
            datetime.datetime(2015, 3, 27, 22, 7, 38, 788861)
            >>> is_closed(step="self-assessment")
            True, "start", datetime.datetime(2014, 3, 27, 22, 7, 38, 788861),
            datetime.datetime(2015, 3, 27, 22, 7, 38, 788861)

        """
        submission_range = (self.submission_start, self.submission_due)
        assessment_ranges = [
            (asmnt.get('start'), asmnt.get('due'))
            for asmnt in self.valid_assessments
        ]

        # Resolve unspecified dates and date strings to datetimes
        start, due, date_ranges = resolve_dates(
            self.start, self.due, [submission_range] + assessment_ranges, self._
        )

        open_range = (start, due)
        assessment_steps = self.assessment_steps
        if step == 'submission':
            open_range = date_ranges[0]
        elif step in assessment_steps:
            step_index = assessment_steps.index(step)
            open_range = date_ranges[1 + step_index]

        # Course staff always have access to the problem
        if course_staff is None:
            course_staff = self.is_course_staff
        if course_staff:
            return False, None, DISTANT_PAST, DISTANT_FUTURE

        if self.date_config_type == DATE_CONFIG_COURSE_END:
            open_range = (
                self.course.start if self.course and self.course.start else DISTANT_PAST,
                self.course.end if self.course and self.course.end else DISTANT_FUTURE
            )
        elif self.date_config_type == DATE_CONFIG_SUBSECTION:
            open_range = (
                self.start if self.start else DISTANT_PAST,
                self.due if self.due else DISTANT_FUTURE
            )

        if self.is_beta_tester:
            beta_start = self._adjust_start_date_for_beta_testers(open_range[0])
            open_range = (beta_start, open_range[1])

        # Check if we are in the open date range
        now = dt.datetime.utcnow().replace(tzinfo=pytz.utc)

        if now < open_range[0]:
            return True, "start", open_range[0], open_range[1]
        elif now >= open_range[1]:
            return True, "due", open_range[0], open_range[1]
        return False, None, open_range[0], open_range[1]

    def get_waiting_details(self, status_details):
        """
        Returns waiting status (boolean value) based on the given status_details.

        Args:
            status_details (dict): A dictionary containing the details of each
                assessment module status. This will contain keys such as
                "peer", "ai", and "staff", referring to dictionaries, which in
                turn will have the key "graded". If this key has a value set,
                these assessment modules have been graded.

        Returns:
            True if waiting for a grade from peer, ai, or staff assessment, else False.

        Examples:
            >>> now = dt.datetime.utcnow().replace(tzinfo=pytz.utc)
            >>> status_details = {
            >>>     'peer': {
            >>>         'completed': None,
            >>>         'graded': now
            >>>     },
            >>>     'ai': {
            >>>         'completed': now,
            >>>         'graded': None
            >>>     }
            >>> }
            >>> self.get_waiting_details(status_details)
            True
        """
        steps = ["peer", "ai", "staff"]  # These are the steps that can be submitter-complete, but lack a grade
        for step in steps:
            if step in status_details and not status_details[step]["graded"]:
                return True
        return False

    def is_released(self, step=None):
        """
        Check if a question has been released.

        Keyword Arguments:
            step (str): The step in the workflow to check.
                None: check whether the problem as a whole is open.
                "submission": check whether the submission section is open.
                "peer-assessment": check whether the peer-assessment section is open.
                "self-assessment": check whether the self-assessment section is open.

        Returns:
            bool
        """
        # By default, assume that we're published, in case the runtime doesn't support publish date.
        if hasattr(self.runtime, 'modulestore'):
            is_published = self.runtime.modulestore.has_published_version(self)
        else:
            is_published = True
        try:
            is_closed, reason, __, __ = self.is_closed(step=step)  # pylint: disable=redeclared-assigned-name
        except DateValidationError:
            # Workaround so that studio_mixin workflow wil still work in the case that we have invalid dates
            is_closed = False
            reason = ''
        is_released = is_published and (not is_closed or reason == 'due')
        if self.start:
            is_released = is_released and dt.datetime.now(pytz.UTC) > parse_date_value(self.start, self._)
        return is_released

    def get_assessment_module(self, mixin_name):
        """
        Get a configured assessment module by name.

        Args:
            mixin_name (str): The name of the mixin (e.g. "self-assessment" or "peer-assessment")

        Returns:
            dict

        Example:
            >>> self.get_assessment_module('peer-assessment')
            {
                "name": "peer-assessment",
                "start": None,
                "due": None,
                "must_grade": 5,
                "must_be_graded_by": 3,
            }
        """
        for assessment in self.valid_assessments:
            if assessment["name"] == mixin_name:
                return assessment
        return None

    def publish_assessment_event(self, event_name, assessment, **kwargs):
        """
        Emit an analytics event for the peer assessment.

        Args:
            event_name (str): An identifier for this event type.
            assessment (dict): The serialized assessment model.

        Returns:
            None

        """
        parts_list = []
        for part in assessment["parts"]:
            # Some assessment parts do not include point values,
            # only written feedback.  In this case, the assessment
            # part won't have an associated option.
            option_dict = None
            if part["option"] is not None:
                option_dict = {
                    "name": part["option"]["name"],
                    "points": part["option"]["points"],
                }

            # All assessment parts are associated with criteria
            criterion_dict = {
                "name": part["criterion"]["name"],
                "points_possible": part["criterion"]["points_possible"]
            }

            parts_list.append({
                "option": option_dict,
                "criterion": criterion_dict,
                "feedback": part["feedback"]
            })

        event_data = {
            "feedback": assessment["feedback"],
            "rubric": {
                "content_hash": assessment["rubric"]["content_hash"],
            },
            "scorer_id": assessment["scorer_id"],
            "score_type": assessment["score_type"],
            "scored_at": assessment["scored_at"],
            "submission_uuid": assessment["submission_uuid"],
            "parts": parts_list
        }

        for key, value in kwargs.items():
            event_data[key] = value

        self.runtime.publish(
            self, event_name,
            event_data
        )

    @XBlock.json_handler
    def publish_event(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Publish the given data to an event.

        Expects key 'event_name' to be present in the data dictionary.
        """

        try:
            event_name = data['event_name']
        except KeyError:
            logger.exception("Could not find the name of the event to be triggered.")
            return {'success': False}

        # Remove the name so we don't publish as part of the data.
        del data['event_name']

        self.runtime.publish(self, event_name, data)
        return {'success': True}

    def get_real_user(self, anonymous_user_id):
        """
        Return the user associated with anonymous_user_id
        Args:
            anonymous_user_id (str): the anonymous user id of the user

        Returns: the user model for the user if it can be identified.
            If the xblock service to converts to a real user fails,
            returns None and logs the error.

        """
        if hasattr(self, "xmodule_runtime"):
            if self.xmodule_runtime.get_real_user is None:  # pylint: disable=no-member
                return None
            user = self.xmodule_runtime.get_real_user(anonymous_user_id)  # pylint: disable=no-member
            if user:
                return user
            logger.exception(
                "XBlock service could not find user for anonymous_user_id '%s'", anonymous_user_id
            )
        return None

    def get_username(self, anonymous_user_id):
        """
        Return the username of the user associated with anonymous_user_id
        Args:
            anonymous_user_id (str): the anonymous user id of the user

        Returns: the username if it can be identified.

        """
        user = self.get_real_user(anonymous_user_id)
        if user:
            return user.username
        return None

    def _adjust_start_date_for_beta_testers(self, start):
        """
        Returns the start date for a Beta tester.
        """
        if days_early_for_beta := getattr(self, 'days_early_for_beta', None):
            delta = dt.timedelta(days_early_for_beta)
            effective = start - delta
            return effective

        return start

    def get_xblock_id(self):
        """
        Returns the xblock id
        """
        return str(self.scope_ids.usage_id)

    def _clean_data(self, data):
        cleaner = Cleaner(tags=set(), strip=True)
        cleaned_text = " ".join(re.split(r"\s+", cleaner.clean(data), flags=re.UNICODE)).strip()
        return cleaned_text

    def index_dictionary(self):
        """
        Return dictionary prepared with module content and type for indexing.
        """

        # return key/value fields in a Python dict object
        # values may be numeric / string or dict
        # default implementation is an empty dict
        xblock_body = super().index_dictionary()

        # Check whether there is only one prompt or more than one
        # If there is single prompt, self.prompt would be simply a string
        # otherwise self.prompt would have json embedded in the string.
        try:
            prompt = {
                f"prompt_{prompt_i}": self._clean_data(prompt.get("description", ""))
                for prompt_i, prompt in enumerate(json.loads(self.prompt))
            }
        except ValueError:
            prompt = {
                "prompt": self._clean_data(self.prompt)
            }

        content = {
            "display_name": self.display_name,
            "title": self.title,
            **prompt
        }

        if "content" in xblock_body:
            xblock_body["content"].update(content)
        else:
            xblock_body["content"] = content

        xblock_body["content_type"] = "ORA"

        return xblock_body



""" Base class for step data collations """
from openassessment.xblock.utils.resolve_dates import DISTANT_FUTURE


class StepDataAPI:
    def __init__(self, block, step=None):
        self._block = block
        problem_closed, closed_reason, start_date, due_date = block.is_closed(step=step)
        self._problem_closed = problem_closed
        self._closed_reason = closed_reason
        self._start_date = start_date
        self._due_date = due_date

    def __repr__(self):
        return "{0}".format(
            {
                "problem_closed": self.problem_closed,
                "closed_reason": self.closed_reason,
                "start_date": self.start_date,
                "due_date": self.due_date,
            }
        )

    @property
    def config_data(self):
        return self._block.api_data.config_data

    @property
    def workflow_data(self):
        return self._block.api_data.workflow_data

    @property
    def problem_closed(self):
        return self._problem_closed

    @property
    def closed_reason(self):
        return self._closed_reason

    @property
    def start_date(self):
        return self._start_date

    @property
    def due_date(self):
        return self._due_date

    @property
    def is_due(self):
        return self.due_date < DISTANT_FUTURE

    @property
    def is_not_available_yet(self):
        return self.problem_closed and self.closed_reason == "start"

    @property
    def is_past_due(self):
        return self.problem_closed and self.closed_reason == "due"



"""
The mixin with handlers for the course ora blocks listing view.

"""

import json

from webob import Response
from xblock.core import XBlock
from openassessment.xblock.staff_area_mixin import require_course_staff


class CourseItemsListingMixin:
    """
    The mixin with handlers for the course ora blocks listing view.

    """

    @XBlock.handler
    @require_course_staff('STAFF_AREA')
    def get_ora2_responses(self, request, suffix=''):  # pylint: disable=unused-argument
        """
        Get information about all ora2 blocks in the course with response count for each step.

        """
        # Import is placed here to avoid model import at project startup.
        from openassessment.data import OraAggregateData
        responses = OraAggregateData.collect_ora2_responses(str(self.course_id))
        return Response(json.dumps(responses), content_type='application/json', charset='UTF-8')



"""
Data Conversion utility methods for handling ORA2 XBlock data transformations and validation.

"""
import json


def convert_training_examples_list_to_dict(examples_list):
    """
    Convert of options selected we store in the problem def,
    which is ordered, to the unordered dictionary of options
    selected that the student training API expects.

    Args:
        examples_list (list): A list of options selected against a rubric.

    Returns:
        A dictionary of the given examples in the list.

    Example:
        >>> examples = [
        >>>     {
        >>>         "answer": {
        >>>             "parts": {
        >>>                 [
        >>>                     {"text:" "Answer part 1"},
        >>>                     {"text:" "Answer part 2"},
        >>>                     {"text:" "Answer part 3"}
        >>>                 ]
        >>>             }
        >>>         },
        >>>         "options_selected": [
        >>>             {
        >>>                 "criterion": "Ideas",
        >>>                 "option": "Fair"
        >>>             },
        >>>             {
        >>>                 "criterion": "Content",
        >>>                 "option": "Good"
        >>>             }
        >>>         ]
        >>>     }
        >>> ]
        >>> convert_training_examples_list_to_dict(examples)
        [
            {
                'answer': {
                    'parts': {
                        [
                            {'text:' 'Answer part 1'},
                            {'text:' 'Answer part 2'},
                            {'text:' 'Answer part 3'}
                        ]
                    }
                 },
                'options_selected': {
                    'Ideas': 'Fair',
                    'Content': 'Good'
                }
            }
        ]

    """
    return [
        {
            'answer': ex['answer'],
            'options_selected': {
                select_dict['criterion']: select_dict['option']
                for select_dict in ex['options_selected']
            }
        }
        for ex in examples_list
    ]


def list_to_conversational_format(str_list):
    """
    String list render method, displaying a list of string values in conversational language.
    ['a'] => 'a';  ['a', 'b'] => 'a and b';  ['a', 'b', 'c'] => 'a, b, and c'

    Args:
        str_list (str[]): List of strings to return in comma-joined/conversational form.

    Returns:
        Combined string.
    """
    if str_list is None:
        return ''
    if len(str_list) < 3:
        return ' and '.join(str_list)
    return '{}, and {}'.format(', '.join(str_list[:-1]), str_list[-1])


def update_saved_response_format(saved_response):
    """
    For backwards compatibility. Initially, problems had only one prompt and a
    string answer. This converts it to the appropriate shape.
    """

    updated_response_format = {}

    try:
        updated_response_format = {
            'answer': json.loads(saved_response),
        }
    except ValueError:
        updated_response_format = {
            'answer': {
                'text': saved_response,
            },
        }

    return updated_response_format


def update_assessments_format(assessments):
    """
    For each example update 'answer' to newer format.

    Args:
        assessments (list): list of assessments
    Returns:
        list of dict
    """
    for assessment in assessments:
        if 'examples' in assessment and assessment['examples']:
            for example in assessment['examples']:
                if isinstance(example, dict) and isinstance(example['answer'], str):
                    example['answer'] = {
                        'parts': [
                            {'text': example['answer']}
                        ]
                    }
                if isinstance(example, dict) and isinstance(example['answer'], list) and example['answer']:
                    example['answer'] = {
                        'parts': [
                            {'text': example_answer} for example_answer in example['answer']
                        ]
                    }
    return assessments


def create_prompts_list(prompt_or_serialized_prompts):
    """
    Construct a list of prompts.

    Initially a block had a single prompt which was saved as a simple string.
    In that case a new prompt dict is constructed from it.

    Args:
        prompt_or_serialized_prompts (unicode): A string which can either
        be a single prompt text or json for a list of prompts.

    Returns:
        list of dict
    """

    if prompt_or_serialized_prompts is None:
        prompt_or_serialized_prompts = ''

    try:
        prompts = json.loads(prompt_or_serialized_prompts)
    except ValueError:
        prompts = [
            {
                'description': prompt_or_serialized_prompts,
            }
        ]
    return prompts


def create_rubric_dict(prompts, criteria):
    """
    Construct a serialized rubric model in the format expected
    by the assessments app.

    Args:
        prompts (list of dict): The rubric prompts.
        criteria (list of dict): The serialized rubric criteria.

    Returns:
        dict

    """
    return {
        "prompts": prompts,
        "criteria": criteria
    }


def clean_criterion_feedback(rubric_criteria, criterion_feedback):
    """
    Remove per-criterion feedback for criteria with feedback disabled
    in the rubric.

    Args:
        rubric_criteria (list): The rubric criteria from the problem definition.
        criterion_feedback (dict): Mapping of criterion names to feedback text.

    Returns:
        dict

    """
    return {
        criterion['name']: criterion_feedback[criterion['name']]
        for criterion in rubric_criteria
        if criterion['name'] in criterion_feedback and criterion.get(
            'feedback', 'disabled'
        ) in ['optional', 'required']
    }


def prepare_submission_for_serialization(submission_data):
    """
    Convert a list of answers into the right format dict for serialization.

    Args:
        submission_data (list of unicode): The answers.

    Returns:
        dict
    """
    return {
        'parts': [{'text': text} for text in submission_data],
    }


def format_files_for_submission(uploaded_files):
    """
    Format a list of FileDescriptors as separate lists of file metadata.

    Args:
    * uploaded_files (List FileDescriptors)

    Returns:
    {
        "file_keys": (List of URLs)
        "files_descriptions": (List of strings)
        "files_names": (List of strings)
        "files_sizes": (List of ints / bytes)
    }
    """

    files_metadata = {}

    for field in ('file_keys', 'files_descriptions', 'files_names', 'files_sizes'):
        files_metadata[field] = []

    for upload in uploaded_files:
        files_metadata['file_keys'].append(upload.key)
        files_metadata['files_descriptions'].append(upload.description)
        files_metadata['files_names'].append(upload.name)
        files_metadata['files_sizes'].append(upload.size)

    return files_metadata


def create_submission_dict(submission, prompts):
    """
    1. Convert from legacy format.
    2. Add prompts to submission['answer']['parts'] to simplify iteration in the template.

    Args:
        submission (dict): Submission dictionary.
        prompts (list of dict): The prompts from the problem definition.

    Returns:
        dict
    """
    parts = [{'prompt': prompt, 'text': ''} for prompt in prompts]

    if 'text' in submission['answer']:
        parts[0]['text'] = submission['answer'].pop('text')
    else:
        for index, part in enumerate(submission['answer'].pop('parts')):
            parts[index]['text'] = part['text']

    submission['answer']['parts'] = parts

    return submission


def make_django_template_key(key):
    """
    Django templates access dictionary items using dot notation,
    which means that dictionary keys with hyphens don't work.
    This function sanitizes a key for use in Django templates
    by replacing hyphens with underscores.

    Args:
        key (basestring): The key to sanitize.

    Returns:
        basestring
    """
    return key.replace('-', '_')


def verify_assessment_parameters(func):
    """
    Verify that the wrapped function receives the given parameters.

    Used for the staff_assess, self_assess, peer_assess functions and uses their data types.

    Args:
        func - the function to be modified

    Returns:
        the modified function
    """
    def verify_and_call(instance, data, suffix):
        """ Inner Method. """
        # Validate the request
        if 'options_selected' not in data:
            return {'success': False, 'msg': instance._('You must provide options selected in the assessment.')}

        if 'overall_feedback' not in data:
            return {'success': False, 'msg': instance._('You must provide overall feedback in the assessment.')}

        if 'criterion_feedback' not in data:
            return {'success': False, 'msg': instance._('You must provide feedback for criteria in the assessment.')}

        return func(instance, data, suffix)
    return verify_and_call



""" Mixin for functionality around the reuse of rubrics between ORAs """
import logging

from opaque_keys import InvalidKeyError
from opaque_keys.edx.keys import UsageKey
from xblock.core import XBlock

logger = logging.getLogger(__name__)


class RubricReuseMixin:

    def _get_course_ora_blocks(self):
        """
        Get all ORA blocks from the course.
        Both draft and published versions are queried, with preference to draft versions
        """
        if hasattr(self.runtime, 'modulestore'):
            return self.runtime.modulestore.get_items(
                self.location.course_key,
                qualifiers={'category': 'openassessment'},
            )
        else:
            logger.info(
                "_get_course_ora_blocks was called for %s but no modulestore was available",
                str(self.location)
            )
            return []

    def get_other_course_ora_blocks(self):
        """
        Returns a list of all ORA blocks in the course, excluding `self`, and any orphaned blocks.
        """
        blocks = self._get_course_ora_blocks()
        self_location = self.location.for_branch(None)

        def not_orphan_or_self(block):
            return block.parent is not None and block.location.for_branch(None) != self_location

        return [block for block in blocks if not_orphan_or_self(block)]

    def get_other_ora_blocks_for_rubric_editor_context(self):
        """
        Return a list of all other openassessment blocks in the course, in the format:
        {
            'display_name': <block display name>
            'location': <block location, as a string>
        }
        """
        other_blocks = self.get_other_course_ora_blocks()
        return [
            {
                'display_name': block.display_name,
                'location': str(block.location.for_branch(None))
            } for block in other_blocks
        ]

    @XBlock.json_handler
    def get_rubric(self, data, suffix=''):  # pylint: disable=unused-argument
        target_block_id = data.get('target_rubric_block_id')
        if target_block_id is None:
            return {'success': False, 'msg': self._('You must specify a block id from which to copy a rubric.')}
        try:
            target_ora_block_locator = UsageKey.from_string(target_block_id)
        except InvalidKeyError:
            return {'success': False, 'msg': self._('Invalid block id.')}
        try:
            rubric = self._get_rubric(target_ora_block_locator)
        except (TargetBlockNotORAException, TargetORABlockNotFoundException):
            return {
                'success': False,
                'msg': self._('No Open Response Assessment found in {course_id} with block id {block_id}').format(
                    course_id=self.location.course_key,
                    block_id=target_block_id
                )
            }
        return {
            'success': True,
            'rubric': {**rubric}
        }

    def _get_rubric(self, target_ora_block_locator):
        target_block = self._get_ora_block(target_ora_block_locator)
        if target_block.category != 'openassessment':
            logger.warning(
                "Requested rubric from %s which is %s not openassessment",
                target_ora_block_locator,
                target_block.category
            )
            raise TargetBlockNotORAException
        return {
            'criteria': target_block.rubric_criteria,
            'feedback_prompt': target_block.rubric_feedback_prompt,
            'feedback_default_text': target_block.rubric_feedback_default_text,
        }

    def _get_ora_block(self, target_ora_block_locator):
        if not hasattr(self.runtime, 'modulestore'):
            logger.warning(
                "Unable to lookup %s for rubric reuse because modulestore is not provided by the runtime",
                target_ora_block_locator
            )
            raise TargetORABlockNotFoundException

        try:
            return self.runtime.modulestore.get_item(target_ora_block_locator)
        except Exception as e:
            logger.warning("Unable to lookup %s for rubric reuse", target_ora_block_locator, exc_info=1)
            raise TargetORABlockNotFoundException from e


class TargetBlockNotORAException(Exception):
    """ Exception raised when _get_rubric is called with the location of a block that isn't an ORA """


class TargetORABlockNotFoundException(Exception):
    """ Exception raised when _get_rubric is called with the location of a block can't be found. """



"""
Tests for data layer of ORA XBlock
"""

from unittest.mock import MagicMock

import ddt


from openassessment.xblock.ui_mixins.mfe.serializers import (
    AssessmentStepsSerializer,
    LeaderboardConfigSerializer,
    RubricConfigSerializer,
    SubmissionConfigSerializer,
)
from openassessment.xblock.test.base import XBlockHandlerTestCase, scenario


class TestSubmissionConfigSerializer(XBlockHandlerTestCase):
    """
    Test for SubmissionConfigSerializer
    """

    def _enable_team_ora(self, xblock):
        """Utility function for mocking team dependencies on the passed xblock"""
        xblock.is_team_assignment = MagicMock(return_value=True)

        xblock.teamset_config = MagicMock()
        xblock.teamset_config.name = xblock.selected_teamset_id

    @scenario("data/submission_open.xml")
    def test_dates(self, xblock):
        # Given an individual (non-teams) ORA
        xblock.teamset_config = MagicMock(return_value=None)

        # When I ask for the submission config
        submission_config = SubmissionConfigSerializer(xblock).data

        # Then I get the expected values
        expected_start = xblock.submission_start
        expected_due = xblock.submission_due
        self.assertEqual(submission_config["startDatetime"], expected_start)
        self.assertEqual(submission_config["endDatetime"], expected_due)

    @scenario("data/basic_scenario.xml")
    def test_dates_missing(self, xblock):
        # Given an individual (non-teams) ORA
        xblock.teamset_config = MagicMock(return_value=None)

        # When I ask for submission config
        submission_config = SubmissionConfigSerializer(xblock).data

        # Then I get the expected values
        self.assertIsNone(submission_config["startDatetime"])
        self.assertIsNone(submission_config["endDatetime"])

    @scenario("data/basic_scenario.xml")
    def test_text_response_config(self, xblock):
        # Given an individual (non-teams) ORA with a text response
        xblock.teamset_config = MagicMock(return_value=None)

        # When I ask for text response config
        submission_config = SubmissionConfigSerializer(xblock).data
        text_response_config = submission_config["textResponseConfig"]

        # Then I get the expected values
        self.assertTrue(text_response_config["enabled"])
        self.assertTrue(text_response_config["required"])
        self.assertEqual(text_response_config["editorType"], "text")
        self.assertFalse(text_response_config["allowLatexPreview"])

    @scenario("data/basic_scenario.xml")
    def test_html_response_config(self, xblock):
        # Given an individual (non-teams) ORA with an html response
        xblock.teamset_config = MagicMock(return_value=None)
        xblock.text_response_editor = "html"

        # When I ask for text response config
        submission_config = SubmissionConfigSerializer(xblock).data
        text_response_config = submission_config["textResponseConfig"]

        # Then I get the expected values
        self.assertEqual(text_response_config["editorType"], "html")

    @scenario("data/basic_scenario.xml")
    def test_latex_preview(self, xblock):
        # Given an individual (non-teams) ORA
        xblock.teamset_config = MagicMock(return_value=None)
        # ... with latex preview enabled
        xblock.allow_latex = True

        # When I ask for text response config
        submission_config = SubmissionConfigSerializer(xblock).data
        text_response_config = submission_config["textResponseConfig"]

        # Then I get the expected values
        self.assertTrue(text_response_config["allowLatexPreview"])

    @scenario("data/file_upload_scenario.xml")
    def test_file_response_config(self, xblock):
        # Given an individual (non-teams) ORA with file upload enabled
        xblock.teamset_config = MagicMock(return_value=None)

        # When I ask for file upload config
        submission_config = SubmissionConfigSerializer(xblock).data
        file_response_config = submission_config["fileResponseConfig"]

        # Then I get the expected values
        self.assertTrue(file_response_config["enabled"])
        self.assertEqual(
            file_response_config["fileUploadLimit"], xblock.MAX_FILES_COUNT
        )
        self.assertEqual(
            file_response_config["fileTypeDescription"],
            xblock.file_upload_type,
        )
        self.assertEqual(
            file_response_config["allowedExtensions"],
            xblock.get_allowed_file_types_or_preset(),
        )
        self.assertEqual(
            file_response_config["blockedExtensions"], xblock.FILE_EXT_BLACK_LIST
        )

    @scenario("data/team_submission.xml")
    def test_team_ora_config(self, xblock):
        # Given a team ORA
        self._enable_team_ora(xblock)

        # When I ask for teams config
        submission_config = SubmissionConfigSerializer(xblock).data
        teams_config = submission_config["teamsConfig"]

        # Then I get the expected values
        self.assertTrue(teams_config["enabled"])
        self.assertEqual(teams_config["teamsetName"], xblock.selected_teamset_id)


@ddt.ddt
class TestRubricConfigSerializer(XBlockHandlerTestCase):
    """
    Test for RubricConfigSerializer
    """

    @ddt.data(True, False)
    @scenario("data/basic_scenario.xml")
    def test_show_during_response(self, xblock, mock_show_rubric):
        # Given a basic setup where I do/not have rubric shown during response
        xblock.show_rubric_during_response = mock_show_rubric

        # When I ask for rubric config
        rubric_config = RubricConfigSerializer(xblock).data

        # Then I get the right values
        self.assertEqual(rubric_config["showDuringResponse"], mock_show_rubric)

    @scenario("data/feedback_only_criterion_staff.xml")
    def test_overall_feedback(self, xblock):
        # Given an ORA block with one criterion

        # When I ask for rubric config
        rubric_config = RubricConfigSerializer(xblock).data

        # Then I get the expected defaults
        criteria = rubric_config["criteria"]
        criterion = criteria[0]
        self.assertEqual(len(criteria), 1)
        self.assertEqual(criterion["name"], "vocabulary")
        self.assertEqual(
            criterion["description"],
            "This criterion accepts only written feedback, so it has no options",
        )

        # ... In this example, feedback is required
        self.assertTrue(criterion["feedbackEnabled"])
        self.assertTrue(criterion["feedbackRequired"])

    @scenario("data/feedback_only_criterion_staff.xml")
    def test_criterion(self, xblock):
        # Given an ORA block with one criterion

        # When I ask for rubric config
        rubric_config = RubricConfigSerializer(xblock).data

        # Then I get the expected defaults
        criteria = rubric_config["criteria"]
        criterion = criteria[0]
        self.assertEqual(len(criteria), 1)
        self.assertEqual(criterion["name"], "vocabulary")
        self.assertEqual(
            criterion["description"],
            "This criterion accepts only written feedback, so it has no options",
        )

        # ... In this example, feedback is required
        self.assertTrue(criterion["feedbackEnabled"])
        self.assertTrue(criterion["feedbackRequired"])

    @scenario("data/feedback_only_criterion_self.xml")
    def test_criterion_disabled_required(self, xblock):
        # Given an ORA block with two criterion

        # When I ask for rubric config
        rubric_config = RubricConfigSerializer(xblock).data

        # Then I get the expected defaults
        criteria = rubric_config["criteria"]

        # .. the first criterion has feedback disabled
        self.assertFalse(criteria[0]["feedbackEnabled"])
        self.assertFalse(criteria[0]["feedbackRequired"])

        # .. the first criterion has feedback required
        self.assertTrue(criteria[1]["feedbackEnabled"])
        self.assertTrue(criteria[1]["feedbackRequired"])

    @scenario("data/file_upload_missing_scenario.xml")
    def test_criterion_optional(self, xblock):
        # Given an ORA block with one criterion, feedback optional

        # When I ask for rubric config
        rubric_config = RubricConfigSerializer(xblock).data

        # Then I get the feedback enabled / required values
        criteria = rubric_config["criteria"]
        criterion = criteria[0]
        self.assertTrue(criterion["feedbackEnabled"])
        self.assertFalse(criterion["feedbackRequired"])

    @scenario("data/basic_scenario.xml")
    def test_criteria(self, xblock):
        # Given an ORA block with multiple criteria
        expected_criteria = xblock.rubric_criteria

        # When I ask for rubric config
        rubric_config = RubricConfigSerializer(xblock).data

        # Then I get the expected number of criteria
        criteria = rubric_config["criteria"]
        self.assertEqual(len(criteria), len(expected_criteria))

    @scenario("data/basic_scenario.xml")
    def test_feedback_config(self, xblock):
        # Given an ORA block with feedback
        xblock.rubric_feedback_prompt = "foo"
        xblock.rubric_feedback_default_text = "bar"

        # When I ask for rubric config
        feedback_config = RubricConfigSerializer(xblock).data["feedbackConfig"]

        # Then I get the expected defaults
        self.assertEqual(feedback_config["description"], xblock.rubric_feedback_prompt)
        self.assertEqual(
            feedback_config["defaultText"], xblock.rubric_feedback_default_text
        )


class TestAssessmentStepsSerializer(XBlockHandlerTestCase):
    """
    Test for AssessmentStepsSerializer
    """

    @scenario("data/basic_scenario.xml")
    def test_order(self, xblock):
        # Given a basic setup
        expected_order = ["peer-assessment", "self-assessment"]
        expected_step_keys = {"training", "peer", "self", "staff"}

        # When I ask for assessment step config
        steps_config = AssessmentStepsSerializer(xblock).data

        # Then I get the right ordering and step keys
        self.assertListEqual(steps_config["order"], expected_order)
        steps = set(steps_config["settings"].keys())
        self.assertSetEqual(steps, expected_step_keys)


class TestPeerSettingsSerializer(XBlockHandlerTestCase):
    """Tests for PeerSettingsSerializer"""

    step_config_key = "peer"

    @scenario("data/basic_scenario.xml")
    def test_peer_settings(self, xblock):
        # Given a basic setup
        expected_must_grade = 5
        expected_grade_by = 3

        # When I ask for peer step config
        peer_config = AssessmentStepsSerializer(xblock).data["settings"][
            self.step_config_key
        ]

        # Then I get the right config
        self.assertEqual(peer_config["minNumberToGrade"], expected_must_grade)
        self.assertEqual(peer_config["minNumberToBeGradedBy"], expected_grade_by)

    @scenario("data/dates_scenario.xml")
    def test_peer_dates(self, xblock):
        # Given a basic setup
        expected_start = "2015-01-02T00:00:00"
        expected_due = "2015-04-01T00:00:00"

        # When I ask for peer step config
        peer_config = AssessmentStepsSerializer(xblock).data["settings"][
            self.step_config_key
        ]

        # Then I get the right dates
        self.assertEqual(peer_config["startTime"], expected_start)
        self.assertEqual(peer_config["endTime"], expected_due)

    @scenario("data/peer_assessment_flex_grading_scenario.xml")
    def test_flex_grading(self, xblock):
        # Given a peer step with flex grading

        # When I ask for peer step config
        peer_config = AssessmentStepsSerializer(xblock).data["settings"][
            self.step_config_key
        ]

        # Then I get the right steps and ordering
        self.assertTrue(peer_config["enableFlexibleGrading"])


class TestTrainingSettingsSerializer(XBlockHandlerTestCase):
    """
    Test for TrainingSettingsSerializer
    """

    step_config_key = "training"

    @scenario("data/student_training.xml")
    def test_enabled(self, xblock):
        # Given an ORA with a training step
        # When I ask for step config
        step_config = AssessmentStepsSerializer(xblock).data["settings"][
            self.step_config_key
        ]

        # Then I get the right config
        self.assertTrue(step_config["required"])

    @scenario("data/basic_scenario.xml")
    def test_disabled(self, xblock):
        # Given an ORA without a training step
        # When I ask for step config
        step_config = AssessmentStepsSerializer(xblock).data["settings"][
            self.step_config_key
        ]

        # Then I get the right config
        self.assertFalse(step_config["required"])


class TestSelfSettingsSerializer(XBlockHandlerTestCase):
    """
    Test for SelfSettingsSerializer
    """

    step_config_key = "self"

    @scenario("data/self_assessment_scenario.xml")
    def test_enabled(self, xblock):
        # Given an ORA with a self assessment step
        # When I ask for step config
        step_config = AssessmentStepsSerializer(xblock).data["settings"][
            self.step_config_key
        ]

        # Then I get the right config
        self.assertTrue(step_config["required"])

    @scenario("data/peer_only_scenario.xml")
    def test_disabled(self, xblock):
        # Given an ORA without a self assessment step
        # When I ask for step config
        step_config = AssessmentStepsSerializer(xblock).data["settings"][
            self.step_config_key
        ]

        # Then I get the right config
        self.assertFalse(step_config["required"])


class TestStaffSettingsSerializer(XBlockHandlerTestCase):
    """
    Test for StaffSettingsSerializer
    """

    step_config_key = "staff"

    @scenario("data/staff_grade_scenario.xml")
    def test_enabled(self, xblock):
        # Given an ORA with a staff assessment step
        # When I ask for step config
        step_config = AssessmentStepsSerializer(xblock).data["settings"][
            self.step_config_key
        ]

        # Then I get the right config
        self.assertTrue(step_config["required"])

    @scenario("data/peer_only_scenario.xml")
    def test_disabled(self, xblock):
        # Given an ORA without a staff assessment step
        # When I ask for step config
        step_config = AssessmentStepsSerializer(xblock).data["settings"][
            self.step_config_key
        ]

        # Then I get the right config
        self.assertFalse(step_config["required"])


class TestLeaderboardConfigSerializer(XBlockHandlerTestCase):
    """
    Test for LeaderboardConfigSerializer
    """

    @scenario("data/leaderboard_show.xml")
    def test_leaderboard(self, xblock):
        # Given I have a leaderboard configured
        number_to_show = xblock.leaderboard_show

        # When I ask for leaderboard config
        leaderboard_config = LeaderboardConfigSerializer(xblock).data

        # Then I get the expected config
        self.assertTrue(leaderboard_config["enabled"])
        self.assertEqual(leaderboard_config["numberOfEntries"], number_to_show)

    @scenario("data/basic_scenario.xml")
    def test_no_leaderboard(self, xblock):
        # Given I don't have a leaderboard configured
        # When I ask for leaderboard config
        leaderboard_config = LeaderboardConfigSerializer(xblock).data

        # Then I get the expected config
        self.assertFalse(leaderboard_config["enabled"])
        self.assertEqual(leaderboard_config["numberOfEntries"], 0)



"""
A mixin for staff grading.
"""

import logging

from openassessment.assessment.api import (
    staff as staff_api,
)

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def staff_assessment_exists(submission_uuid):
    """
    Returns True if there exists a staff assessment for the given uuid. False otherwise.
    """
    return staff_api.get_latest_staff_assessment(submission_uuid) is not None


def staff_path_and_context(api_data):
    """
    Retrieve the correct template path and template context for the handler to render.
    """
    return "openassessmentblock/staff/oa_staff_grade.html", staff_context(api_data)


def render_staff_assessment(api_data):
    """
    Renders the Staff Assessment HTML section of the XBlock
    Generates the staff assessment HTML for the Open
    Assessment XBlock. See OpenAssessmentBlock.render_assessment() for
    more information on rendering XBlock sections.
    """
    path, context_dict = staff_path_and_context(api_data)
    return api_data.config_data.render_assessment(path, context_dict)


def staff_context(api_data):
    """
    Retrieve the correct template path and template context for the handler to render.
    """
    step_data = api_data.staff_assessment_data
    translate = api_data.config_data.translate

    not_available_context = {
        "status_value": translate("Not Available"),
        "button_active": "disabled=disabled aria-expanded=false",
        "step_classes": "is--unavailable",
    }

    if step_data.is_cancelled:
        context = {
            "status_value": translate("Cancelled"),
            "icon_class": "fa-exclamation-triangle",
            "step_classes": "is--unavailable",
            "button_active": "disabled=disabled aria-expanded=false",
        }
    elif step_data.is_done:  # Staff grade exists and all steps completed.
        context = {
            "status_value": translate("Complete"),
            "icon_class": "fa-check",
            "step_classes": "is--complete is--empty",
            "button_active": "disabled=disabled aria-expanded=false",
        }
    elif step_data.is_waiting:
        # If we are in the 'waiting' workflow, this means that a staff grade cannot exist
        # (because if a staff grade did exist, we would be in 'done' regardless of whether other
        # peers have assessed). Therefore we show that we are waiting on staff to provide a grade.
        context = {
            "status_value": translate("Not Available"),
            "message_title": translate("Waiting for a Staff Grade"),
            "message_content": translate(
                "Check back later to see if a course staff member has assessed "
                "your response. You will receive your grade after the assessment "
                "is complete."
            ),
            "step_classes": "is--showing",
            "button_active": "aria-expanded=true",
        }
    elif not step_data.has_status:
        context = not_available_context
    else:  # status is 'self' or 'peer', indicating that the student still has work to do.
        if staff_assessment_exists(api_data.workflow_data.submission_uuid):
            context = {
                "status_value": translate("Complete"),
                "icon_class": "fa-check",
                "message_title": translate("You Must Complete the Steps Above to View Your Grade"),
                "message_content": translate(
                    "Although a course staff member has assessed your response, "
                    "you will receive your grade only after you have completed "
                    "all the required steps of this problem."
                ),
                "step_classes": "is--initially--collapsed",
                "button_active": "aria-expanded=false",
            }
        else:  # Both student and staff still have work to do, just show "Not Available".
            context = not_available_context

    context["xblock_id"] = api_data.config_data.get_xblock_id()
    return context



# pylint: skip-file


from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Assessment',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('submission_uuid', models.CharField(max_length=128, db_index=True)),
                ('scored_at', models.DateTimeField(default=django.utils.timezone.now, db_index=True)),
                ('scorer_id', models.CharField(max_length=40, db_index=True)),
                ('score_type', models.CharField(max_length=2)),
                ('feedback', models.TextField(default='', max_length=10000, blank=True)),
            ],
            options={
                'ordering': ['-scored_at', '-id'],
            },
        ),
        migrations.CreateModel(
            name='AssessmentFeedback',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('submission_uuid', models.CharField(unique=True, max_length=128, db_index=True)),
                ('feedback_text', models.TextField(default='', max_length=10000)),
                ('assessments', models.ManyToManyField(default=None, related_name='assessment_feedback', to='assessment.Assessment')),
            ],
        ),
        migrations.CreateModel(
            name='AssessmentFeedbackOption',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('text', models.CharField(unique=True, max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='AssessmentPart',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('feedback', models.TextField(default='', blank=True)),
                ('assessment', models.ForeignKey(related_name='parts', to='assessment.Assessment', on_delete=models.CASCADE)),
            ],
        ),
        migrations.CreateModel(
            name='Criterion',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=100)),
                ('label', models.CharField(max_length=100, blank=True)),
                ('order_num', models.PositiveIntegerField()),
                ('prompt', models.TextField(max_length=10000)),
            ],
            options={
                'ordering': ['rubric', 'order_num'],
            },
        ),
        migrations.CreateModel(
            name='CriterionOption',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('order_num', models.PositiveIntegerField()),
                ('points', models.PositiveIntegerField()),
                ('name', models.CharField(max_length=100)),
                ('label', models.CharField(max_length=100, blank=True)),
                ('explanation', models.TextField(max_length=10000, blank=True)),
                ('criterion', models.ForeignKey(related_name='options', to='assessment.Criterion', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['criterion', 'order_num'],
            },
        ),
        migrations.CreateModel(
            name='PeerWorkflow',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('student_id', models.CharField(max_length=40, db_index=True)),
                ('item_id', models.CharField(max_length=128, db_index=True)),
                ('course_id', models.CharField(max_length=40, db_index=True)),
                ('submission_uuid', models.CharField(unique=True, max_length=128, db_index=True)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, db_index=True)),
                ('completed_at', models.DateTimeField(null=True, db_index=True)),
                ('grading_completed_at', models.DateTimeField(null=True, db_index=True)),
                ('cancelled_at', models.DateTimeField(null=True, db_index=True)),
            ],
            options={
                'ordering': ['created_at', 'id'],
            },
        ),
        migrations.CreateModel(
            name='PeerWorkflowItem',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('submission_uuid', models.CharField(max_length=128, db_index=True)),
                ('started_at', models.DateTimeField(default=django.utils.timezone.now, db_index=True)),
                ('scored', models.BooleanField(default=False)),
                ('assessment', models.ForeignKey(to='assessment.Assessment', null=True, on_delete=models.CASCADE)),
                ('author', models.ForeignKey(related_name='graded_by', to='assessment.PeerWorkflow', on_delete=models.CASCADE)),
                ('scorer', models.ForeignKey(related_name='graded', to='assessment.PeerWorkflow', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['started_at', 'id'],
            },
        ),
        migrations.CreateModel(
            name='Rubric',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('content_hash', models.CharField(unique=True, max_length=40, db_index=True)),
                ('structure_hash', models.CharField(max_length=40, db_index=True)),
            ],
        ),
        migrations.CreateModel(
            name='StudentTrainingWorkflow',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('submission_uuid', models.CharField(unique=True, max_length=128, db_index=True)),
                ('student_id', models.CharField(max_length=40, db_index=True)),
                ('item_id', models.CharField(max_length=128, db_index=True)),
                ('course_id', models.CharField(max_length=40, db_index=True)),
            ],
        ),
        migrations.CreateModel(
            name='StudentTrainingWorkflowItem',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('order_num', models.PositiveIntegerField()),
                ('started_at', models.DateTimeField(auto_now_add=True)),
                ('completed_at', models.DateTimeField(default=None, null=True)),
            ],
            options={
                'ordering': ['workflow', 'order_num'],
            },
        ),
        migrations.CreateModel(
            name='TrainingExample',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('raw_answer', models.TextField(blank=True)),
                ('content_hash', models.CharField(unique=True, max_length=40, db_index=True)),
                ('options_selected', models.ManyToManyField(to='assessment.CriterionOption')),
                ('rubric', models.ForeignKey(to='assessment.Rubric', on_delete=models.CASCADE)),
            ],
        ),
        migrations.AddField(
            model_name='studenttrainingworkflowitem',
            name='training_example',
            field=models.ForeignKey(to='assessment.TrainingExample', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='studenttrainingworkflowitem',
            name='workflow',
            field=models.ForeignKey(related_name='items', to='assessment.StudentTrainingWorkflow', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='criterion',
            name='rubric',
            field=models.ForeignKey(related_name='criteria', to='assessment.Rubric', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='assessmentpart',
            name='criterion',
            field=models.ForeignKey(related_name='+', to='assessment.Criterion', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='assessmentpart',
            name='option',
            field=models.ForeignKey(related_name='+', to='assessment.CriterionOption', null=True, on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='assessmentfeedback',
            name='options',
            field=models.ManyToManyField(default=None, related_name='assessment_feedback', to='assessment.AssessmentFeedbackOption'),
        ),
        migrations.AddField(
            model_name='assessment',
            name='rubric',
            field=models.ForeignKey(to='assessment.Rubric', on_delete=models.CASCADE),
        ),
        migrations.AlterUniqueTogether(
            name='studenttrainingworkflowitem',
            unique_together={('workflow', 'order_num')},
        ),
    ]



""" A mixin for self Assessments. """
import logging

from openassessment.assessment.api import self as self_api
from openassessment.workflow import api as workflow_api

from openassessment.xblock.utils.data_conversion import (
    clean_criterion_feedback,
    create_rubric_dict,
)

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

messages = {
    "must_submit": "You must submit a response before you can perform a self-assessment.",
    "could_not_submit": "Your self assessment could not be submitted.",
}


def self_assess(api_data, data, suffix=""):  # pylint: disable=unused-argument
    """
    Create a self-assessment for a submission.

    Args:
        data (dict): Must have the following keys:
            options_selected (dict): Dictionary mapping criterion names to option values.

    Returns:
        Dict with keys "success" (bool) indicating success/failure
        and "msg" (unicode) containing additional information if an error occurs.
    """
    # Import is placed here to avoid model import at project startup.
    translate = api_data.config_data.translate
    submission_uuid = api_data.submission_data.submission_uuid

    def failure_response(reason):
        return {"success": False, "msg": translate(reason)}

    if submission_uuid is None:
        return failure_response(messages["must_submit"])

    step_data = api_data.self_assessment_data
    try:
        assessment = self_api.create_assessment(
            step_data.submission_uuid,
            step_data.student_item_dict["student_id"],
            data["options_selected"],
            clean_criterion_feedback(step_data.rubric_criteria, data["criterion_feedback"]),
            data["overall_feedback"],
            create_rubric_dict(step_data.prompts, step_data.rubric_criteria_with_labels),
        )
        api_data.config_data.publish_assessment_event("openassessmentblock.self_assess", assessment)
        # After we've created the self-assessment, we need to update the workflow.
        api_data.workflow_data.update_workflow_status()
    except (
        self_api.SelfAssessmentRequestError,
        workflow_api.AssessmentWorkflowRequestError,
    ):
        logger.warning(
            "An error occurred while submitting a self assessment for the submission %s",
            submission_uuid,
            exc_info=True,
        )
        return failure_response(messages["could_not_submit"])
    except (
        self_api.SelfAssessmentInternalError,
        workflow_api.AssessmentWorkflowInternalError,
    ):
        logger.exception(
            "An error occurred while submitting a self assessment for the submission %s",
            submission_uuid,
        )
        return failure_response(messages["could_not_submit"])
    else:
        return {"success": True, "msg": ""}



"""
Actions for file management
"""
import logging
import os
from openassessment.fileupload.exceptions import FileUploadError

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def save_files_descriptions(block_config, submission_info, data):
    """
    Save the metadata for each uploaded file.

    Args:
        block_config (ORAConfigAPI)
        submission_info (SubmissionAPI)
        data (dict): Data should have a single key 'fileMetadata' that contains
            a list of dictionaries with the following keys: 'description','fileName', and 'fileSize'
        each element of the list maps to a single file

    Returns:
        dict: Contains a bool 'success' and unicode string 'msg'.
    """
    failure_response = {
        "success": False,
        "msg": block_config.translate("Files descriptions were not submitted."),
    }

    if "fileMetadata" not in data:
        return failure_response

    if not isinstance(data["fileMetadata"], list):
        return failure_response

    file_data = [
        {
            "description": item["description"],
            "name": item["fileName"],
            "size": item["fileSize"],
        }
        for item in data["fileMetadata"]
    ]

    for new_upload in file_data:
        if not all(
            [
                isinstance(new_upload["description"], str),
                isinstance(new_upload["name"], str),
                isinstance(new_upload["size"], int),
            ]
        ):
            return failure_response

    try:
        submission_info.files.file_manager.append_uploads(*file_data)
        # Emit analytics event...
        block_config.publish_event(
            "openassessmentblock.save_files_descriptions",
            {"saved_response": submission_info.saved_files_descriptions},
        )
    except FileUploadError as exc:
        logger.exception(
            "FileUploadError: file description for data %s failed with error %s",
            data,
            exc,
            exc_info=True,
        )
        return {
            "success": False,
            "msg": block_config._("Files metadata could not be saved."),
        }
    except Exception as exc:  # pylint: disable=broad-except
        logger.exception(
            "FileUploadError: unhandled exception for data %s. Error: %s",
            data,
            exc,
            exc_info=True,
        )
    return {"success": True, "msg": ""}


def upload_url(block_config, submission_info, data):
    """
    Request a URL to be used for uploading content related to this
    submission.

    Returns:
        A URL to be used to upload content associated with this submission.

    """
    if "contentType" not in data or "filename" not in data:
        return {
            "success": False,
            "msg": block_config.translate("There was an error uploading your file."),
        }

    if not block_config.allow_multiple_files:
        # Here we check if there are existing file uploads by checking for
        # an existing download url for any of the upload slots.
        # Note that we can't use self.saved_files_descriptions because that
        # is populated before files are uploaded
        for file_index in range(submission_info.files.max_allowed_uploads):
            file_url = submission_info.files.get_download_url(file_index)
            if file_url:
                return {
                    "success": False,
                    "msg": block_config.translate("Only a single file upload is allowed for this assessment."),
                }

    file_num = int(data.get("filenum", 0))

    _, file_ext = os.path.splitext(data["filename"])
    file_ext = file_ext.strip(".") if file_ext else None
    content_type = data["contentType"]

    # Validate that there are no data issues and file type is allowed
    if not submission_info.files.is_supported_upload_type(file_ext, content_type):
        return {
            "success": False,
            "msg": block_config.translate(
                "File upload failed: unsupported file type."
                "Only the supported file types can be uploaded."
                "If you have questions, please reach out to the course team."
            ),
        }

    # Attempt to upload
    file_num = int(data.get("filenum", 0))
    try:
        key = submission_info.files.get_file_key(file_num)
        url = submission_info.files.get_upload_url(key, content_type)
        return {"success": True, "url": url}
    except FileUploadError:
        logger.exception("FileUploadError:Error retrieving upload URL for the data: %s.", data)
        return {
            "success": False,
            "msg": block_config.translate("Error retrieving upload URL."),
        }


def download_url(submission_info, data):
    """
    Request a download URL.

    Returns:
        A URL to be used for downloading content related to the submission.
    """
    file_num = int(data.get("filenum", 0))
    return {"success": True, "url": submission_info.files.get_download_url(file_num)}


def remove_uploaded_file(block_config, submission_info, data):
    """
    Removes uploaded user file.
    """
    file_index = data.get("filenum", -1)
    try:
        file_index = int(file_index)
    except ValueError:
        file_index = -1
    student_item_key = submission_info.files.get_file_key(file_index)
    if submission_info.files.can_delete_file(file_index):
        try:
            submission_info.files.delete_uploaded_file(file_index)

            # Emit analytics event...
            block_config.publish_event(
                "openassessmentblock.remove_uploaded_file",
                {"student_item_key": student_item_key},
            )
            logger.debug("Deleted file %s", student_item_key)
            return {"success": True}
        except FileUploadError as exc:
            logger.exception(
                "FileUploadError: Error when deleting file %s : %s",
                student_item_key,
                exc,
                exc_info=True,
            )
        except Exception as exc:  # pylint: disable=broad-except
            logger.exception(
                "FileUploadError: unhandled exception for data %s. Error: %s",
                data,
                exc,
                exc_info=True,
            )

    return {"success": False}



""" A Mixin for Response submissions. """
import logging

from django.utils.functional import cached_property

from openassessment.fileupload import api as file_upload_api
from openassessment.fileupload.exceptions import FileUploadError

from openassessment.data import OraSubmissionAnswerFactory

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class FilesMixin:
    """
    Files Mixin introducing all Files-related functionality.
    """

    ALLOWED_IMAGE_MIME_TYPES = ['image/gif', 'image/jpeg', 'image/pjpeg', 'image/png']  # pragma: no cover
    ALLOWED_IMAGE_EXTENSIONS = ['gif', 'jpg', 'jpeg', 'jfif', 'pjpeg', 'pjp', 'png']  # pragma: no cover

    ALLOWED_FILE_MIME_TYPES = ['application/pdf'] + ALLOWED_IMAGE_MIME_TYPES  # pragma: no cover
    ALLOWED_FILE_EXTENSIONS = ['pdf'] + ALLOWED_IMAGE_EXTENSIONS  # pragma: no cover

    MAX_FILES_COUNT = 20  # pragma: no cover

    # taken from http://www.howtogeek.com/137270/50-file-extensions-that-are-potentially-dangerous-on-windows/
    # and http://pcsupport.about.com/od/tipstricks/a/execfileext.htm
    # left out .js and office extensions
    FILE_EXT_BLACK_LIST = [
        'exe', 'msi', 'app', 'dmg', 'com', 'pif', 'application', 'gadget',
        'msp', 'scr', 'hta', 'cpl', 'msc', 'jar', 'bat', 'cmd', 'vb', 'vbs',
        'jse', 'ws', 'wsf', 'wsc', 'wsh', 'scf', 'lnk', 'inf', 'reg', 'ps1',
        'ps1xml', 'ps2', 'ps2xml', 'psc1', 'psc2', 'msh', 'msh1', 'msh2', 'mshxml',
        'msh1xml', 'msh2xml', 'action', 'apk', 'app', 'bin', 'command', 'csh',
        'ins', 'inx', 'ipa', 'isu', 'job', 'mst', 'osx', 'out', 'paf', 'prg',
        'rgs', 'run', 'sct', 'shb', 'shs', 'u3p', 'vbscript', 'vbe', 'workflow',
        'htm', 'html',
    ]

    FILE_UPLOAD_PRESETS = {
        'image': {
            'mime_types': ALLOWED_IMAGE_MIME_TYPES,
            'extensions': ALLOWED_IMAGE_EXTENSIONS
        },
        'pdf-and-image': {
            'mime_types': ALLOWED_FILE_MIME_TYPES,
            'extensions': ALLOWED_FILE_EXTENSIONS,
        },
        'custom': {}
    }

    @cached_property
    def file_manager(self):
        return file_upload_api.FileUploadManager(self)

    # FILE UPLOADS
    @classmethod
    def _get_url_by_file_key(cls, key):
        """
        Return download url for some particular file key.

        """
        url = ''
        try:
            if key:
                url = file_upload_api.get_download_url(key)
        except FileUploadError as exc:
            logger.exception(
                "FileUploadError: Download url for file key %s failed with error %s",
                key,
                exc,
                exc_info=True
            )

        return url

    @classmethod
    def get_download_urls_from_submission(cls, submission):
        """
        Returns a download URLs for retrieving content within a submission.

        Args:
            submission (dict): Dictionary containing an answer and a file_keys.
                The file_keys is used to try and retrieve a download urls
                with related content

        Returns:
            List of FileDescriptor dicts for each file associated with the submission

        """
        urls = []
        raw_answer = submission.get('answer')
        answer = OraSubmissionAnswerFactory.parse_submission_raw_answer(raw_answer)
        for file_upload in answer.get_file_uploads(missing_blank=True):
            file_download_url = cls._get_url_by_file_key(file_upload.key)
            if file_download_url:
                urls.append(
                    file_upload_api.FileDescriptor(
                        download_url=file_download_url,
                        description=file_upload.description,
                        name=file_upload.name,
                        size=file_upload.size,
                        show_delete_button=False
                    )._asdict()
                )
        return urls

    def get_files_info_from_user_state(self, username):
        """
        Returns the files information from the user state for a given username.

        If the files information is present in the user state, return a list of following tuple:
        (file_download_url, file_description, file_name)

        Arguments:
            username(str): user's name whose state is being check for files information.
        Returns:
            List of FileDescriptor dicts, if present, else empty list.
        """

        files_info = []
        user_state = self.get_user_state(username)
        item_dict = self.get_student_item_dict_from_username_or_email(username)
        if 'saved_files_descriptions' in user_state:
            # pylint: disable=protected-access
            files_descriptions = file_upload_api._safe_load_json_list(
                user_state.get('saved_files_descriptions'),
                log_error=True
            )
            files_names = file_upload_api._safe_load_json_list(
                user_state.get('saved_files_names', '[]'),
                log_error=True
            )
            for index, description in enumerate(files_descriptions):
                file_key = file_upload_api.get_student_file_key(item_dict, index)
                download_url = self._get_url_by_file_key(file_key)
                if download_url:
                    file_name = files_names[index] if index < len(files_names) else ''
                    files_info.append(
                        file_upload_api.FileDescriptor(
                            download_url=download_url,
                            description=description,
                            name=file_name,
                            size=None,
                            show_delete_button=False
                        )._asdict()
                    )
                else:
                    # If file has been removed, the URL doesn't exist
                    logger.info(
                        "URLWorkaround: no URL for description %s & key %s for user:%s",
                        description,
                        username,
                        file_key
                    )
                    continue
        return files_info

    def get_all_upload_urls_for_user(self, username_or_email):
        """
        For a particular ORA block, get the download URLs for all the files uploaded and still present.

        Used for an extreme edge case, where the stored files indices are out of sync with
        the uploaded files, this is a last resort to get the download URLs of all the files
        that have been uploaded by a learner in an ORA block(and haven't been deleted from the storage).
        Starting from 0 index to maximum file upload count possible, this checks if a file exists against
        every index. If present, add the info, else repeat it for the next indices.

        Arguments:
            username_or_email(str): username or email of the learner whose files' information is to be obtained.
        Returns:
            List of FileDescriptor dicts
        """
        file_uploads = []
        student_item_dict = self.get_student_item_dict_from_username_or_email(username_or_email)
        for index in range(self.MAX_FILES_COUNT):
            file_key = file_upload_api.get_student_file_key(student_item_dict, index)
            download_url = ''
            try:
                download_url = file_upload_api.get_download_url(file_key)
            except FileUploadError:
                pass

            if download_url:
                logger.info(
                    "Download URL exists for key %s in block %s for user %s",
                    file_key,
                    username_or_email,
                    str(self.location)
                )
                file_uploads.append(
                    file_upload_api.FileDescriptor(
                        download_url=download_url,
                        description='',
                        name='',
                        size=None,
                        show_delete_button=False
                    )._asdict()
                )
            else:
                continue

        return file_uploads

    def get_allowed_file_types_or_preset(self):
        """
        If allowed files are not explicitly set for file uploads, use preset extensions
        """
        if self.white_listed_file_types:
            return self.white_listed_file_types
        elif self.file_upload_type == 'image':
            return self.ALLOWED_IMAGE_EXTENSIONS
        elif self.file_upload_type == 'pdf-and-image':
            return self.ALLOWED_FILE_EXTENSIONS
        return None



"""
Errors for self-assessment
"""
from .base import AssessmentError


class SelfAssessmentError(AssessmentError):
    """Generic Self Assessment Error

    Raised when an error occurs while processing a request related to the
    Self Assessment Workflow.

    """


class SelfAssessmentRequestError(SelfAssessmentError):
    """
    There was a problem with the request for a self-assessment.
    """


class SelfAssessmentInternalError(SelfAssessmentError):
    """
    There was an internal problem while accessing the self-assessment api.
    """



"""
Mixin to cover various template and UI related utilities
"""

from django.utils.translation import gettext as _
from web_fragments.fragment import Fragment

from openassessment.xblock.utils.defaults import (
    PEER_ASSESSMENT_MODULES,
    SELF_ASSESSMENT_MODULES,
    SELF_TO_PEER_ASSESSMENT_MODULES,
    SELF_TO_STAFF_ASSESSMENT_MODULES,
    STAFF_ASSESSMENT_MODULES,
)
from openassessment.xblock.utils.editor_config import AVAILABLE_EDITORS
from openassessment.xblock.apis.assessments.staff_assessment_api import StaffAssessmentAPI
from openassessment.xblock.load_static import LoadStatic


UI_MODELS = {
    "submission": {
        "name": "submission",
        "class_id": "step--response",
        "title": "Your Response"
    },
    "student-training": {
        "name": "student-training",
        "class_id": "step--student-training",
        "title": "Learn to Assess Responses"
    },
    "peer-assessment": {
        "name": "peer-assessment",
        "class_id": "step--peer-assessment",
        "title": "Assess Peers"
    },
    "self-assessment": {
        "name": "self-assessment",
        "class_id": "step--self-assessment",
        "title": "Assess Your Response"
    },
    "staff-assessment": {
        "name": "staff-assessment",
        "class_id": "step--staff-assessment",
        "title": "Staff Grade"
    },
    "grade": {
        "name": "grade",
        "class_id": "step--grade",
        "title": "Your Grade:"
    },
    "leaderboard": {
        "name": "leaderboard",
        "class_id": "step--leaderboard",
        "title": "Top Responses"
    }
}


class OpenAssessmentTemplatesMixin:
    """
    This helps to get templates for different type of assessment that is
    offered.
    """

    VALID_ASSESSMENT_TYPES_DISPLAY_NAMES = {
        "peer-assessment": _("Peer Assessment Only"),
        "self-assessment": _("Self Assessment Only"),
        "staff-assessment": _("Staff Assessment Only"),
        "self-to-peer": _("Self Assessment to Peer Assessment"),
        "self-to-staff": _("Self Assessment to Staff Assessment")
    }

    VALID_ASSESSMENT_TYPES_ASSESSMENT_MODULE = {
        "self-assessment": SELF_ASSESSMENT_MODULES,
        "peer-assessment": PEER_ASSESSMENT_MODULES,
        "staff-assessment": STAFF_ASSESSMENT_MODULES,
        "self-to-peer": SELF_TO_PEER_ASSESSMENT_MODULES,
        "self-to-staff": SELF_TO_STAFF_ASSESSMENT_MODULES,
    }

    @classmethod
    def templates(cls):
        """
        Returns a list of dictionary field: value objects that describe possible templates.
        """
        templates = []
        for assessment_type, display_name in cls.VALID_ASSESSMENT_TYPES_DISPLAY_NAMES.items():
            template_id = assessment_type
            template = cls._create_template_dict(template_id, display_name)
            templates.append(template)
        return templates

    @classmethod
    def filter_templates(cls, template, _):
        """
        Filters the list of templates for the template view. Filter out only peer-assessment, because peer-assessment
        is the "default" template we use for ORA. Without this, it would display twice.
        See get_component_templates in studio.
        """
        return template['template_id'] != 'peer-assessment'

    @classmethod
    def _create_template_dict(cls, template_id, display_name):
        """
        Creates a dictionary for serving various metadata for the template.

        Args:
            template_id(str): template id of what assessment template needs to be served.
            display_name(str): display name of template.

        Returns:
            A dictionary with proper keys to be consumed.
        """
        return {
            "template_id": template_id,
            "metadata": {
                "display_name": display_name,
            }
        }

    @classmethod
    def get_template(cls, template_id):
        """
        Helps to generate various option level template for ORA.

        Args:
            template_id(str): template id of what assessment template needs to be served.

        Returns:
            A dictionary of payload to be consumed by Studio.
        """
        rubric_assessments = cls._create_rubric_assessment_dict(template_id)
        return {
            "data": rubric_assessments
        }

    @classmethod
    def _create_rubric_assessment_dict(cls, template_id):
        """
        Creates a dictionary of parameters to be passed while creating ORA xblock.

        Args:
            template_id(str): template id of what assessment template needs to be served.

        Returns:
            A dictionary of payload to be consumed by Studio.
        """
        assessment_module = cls.VALID_ASSESSMENT_TYPES_ASSESSMENT_MODULE \
            .get(template_id)
        return {
            "rubric_assessments": assessment_module
        }

    def _create_ui_models(self):
        """Combine UI attributes and XBlock configuration into a UI model.

        This method takes all configuration for this XBlock instance and appends
        UI attributes to create a UI Model for rendering all assessment modules.
        This allows a clean separation of static UI attributes from persistent
        XBlock configuration.

        """
        ui_models = [UI_MODELS["submission"]]
        staff_assessment_required = False
        for assessment in self.valid_assessments:
            if assessment["name"] == "staff-assessment":
                if not assessment["required"]:
                    continue
                staff_assessment_required = True
            ui_model = UI_MODELS.get(assessment["name"])
            if ui_model:
                ui_models.append(dict(assessment, **ui_model))

        if not staff_assessment_required and StaffAssessmentAPI.staff_assessment_exists(self.submission_uuid):
            ui_models.append(UI_MODELS["staff-assessment"])

        ui_models.append(UI_MODELS["grade"])

        if self.leaderboard_show > 0 and not self.teams_enabled:
            ui_models.append(UI_MODELS["leaderboard"])

        return ui_models

    def _create_fragment(
        self,
        template,
        context_dict,
        initialize_js_func,
        additional_css=None,
        additional_js=None,
        additional_js_context=None
    ):
        """
        Creates a fragment for display.

        """
        fragment = Fragment(template.render(context_dict))

        if additional_css is None:
            additional_css = []
        if additional_js is None:
            additional_js = []

        i18n_service = self.runtime.service(self, 'i18n')
        if hasattr(i18n_service, 'get_language_bidi') and i18n_service.get_language_bidi():
            css_url = LoadStatic.get_url("openassessment-rtl.css")
        else:
            css_url = LoadStatic.get_url("openassessment-ltr.css")

        # TODO: load CSS and JavaScript as URLs once they can be served by the CDN
        for css in additional_css:
            fragment.add_css_url(css)
        fragment.add_css_url(css_url)

        # minified additional_js should be already included in 'make javascript'
        fragment.add_javascript_url(LoadStatic.get_url("openassessment-lms.js"))

        js_context_dict = {
            "ALLOWED_IMAGE_MIME_TYPES": self.ALLOWED_IMAGE_MIME_TYPES,
            "ALLOWED_FILE_MIME_TYPES": self.ALLOWED_FILE_MIME_TYPES,
            "FILE_EXT_BLACK_LIST": self.FILE_EXT_BLACK_LIST,
            "FILE_TYPE_WHITE_LIST": self.white_listed_file_types,
            "MAXIMUM_FILE_UPLOAD_COUNT": self.MAX_FILES_COUNT,
            "TEAM_ASSIGNMENT": self.is_team_assignment(),
            "AVAILABLE_EDITORS": AVAILABLE_EDITORS,
            "TEXT_RESPONSE_EDITOR": self.text_response_editor,
        }
        # If there's any additional data to be passed down to JS
        # include it in the context dict
        if additional_js_context:
            js_context_dict.update({"CONTEXT": additional_js_context})

        fragment.initialize_js(initialize_js_func, js_context_dict)
        return fragment



"""
Views related to submissions
"""
import logging

from django.core.exceptions import ObjectDoesNotExist
from xblock.exceptions import NoSuchServiceError

from openassessment.xblock.utils.data_conversion import (
    create_submission_dict,
    list_to_conversational_format,
)
from openassessment.xblock.utils.user_data import get_user_preferences


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def render_submission(config, submission_info):
    """
    Renders the Submission HTML section of the XBlock

    Generates the submission HTML for the first section of an Open
    Assessment XBlock. See OpenAssessmentBlock.render_assessment() for
    more information on rendering XBlock sections.

    Needs to support the following scenarios:
    - Unanswered and Open
    - Unanswered and Closed
    - Saved
    - Saved and Closed
    - Submitted
    - Submitted and Closed
    - Submitted, waiting assessment
    - Submitted and graded
    """
    context = get_submission_context(config, submission_info)
    path = get_submission_path(submission_info)

    return config.render_assessment(path, context_dict=context)


def get_submission_path(submission_info):
    """
    Given info about the submission, return the appropriate template path

    Returns:
    - path (String) to the template
    """

    # Template Paths
    template_dir = "openassessmentblock/response"
    submission_template_paths = {
        "default": "oa_response",
        "closed": "oa_response_closed",
        "unavailable": "oa_response_unavailable",
        "team_already_submitted": "oa_response_team_already_submitted",
        "cancelled": "oa_response_cancelled",
        "graded": "oa_response_graded",
        "submitted": "oa_response_submitted",
    }
    full_paths = {k: f"{template_dir}/{path}.html" for (k, path) in submission_template_paths.items()}

    # Response is unavailable (not yet open or past due date)
    if not submission_info.has_submitted and submission_info.problem_is_inaccessible:
        if submission_info.is_past_due:
            return full_paths["closed"]
        if submission_info.is_not_available_yet:
            return full_paths["unavailable"]

    # Response is unavailable (team assignment where user hasn't submitted and is not on a team)
    elif submission_info.is_team_assignment and submission_info.team_id is None:
        return full_paths["unavailable"]

    # Response not yet submitted
    elif not submission_info.has_submitted:
        if submission_info.is_team_assignment and submission_info.team_previously_submitted_without_student:
            return full_paths["team_already_submitted"]

    # Cancelled: Instructor has cancelled this response
    elif submission_info.has_been_cancelled:
        return full_paths["cancelled"]

    # Done: Submitted and received final grade
    elif submission_info.has_received_final_grade:
        return full_paths["graded"]

    # Submitted and waiting for a grade
    else:
        return full_paths["submitted"]

    return full_paths["default"]


def get_team_submission_context(config):
    """
    Populate the passed context object with team info, including a set of students on
    the team with submissions to the current item from another team, under the key
    `team_members_with_external_submissions`.

    Args:
        config: Access to ORA config API
    Returns
        (dict): context with team-related fields
    """
    from submissions import team_api

    team_context = {}

    try:
        team_info = config.get_team_info()
        if team_info:
            team_context = team_info
            if config.is_course_staff:
                return team_context
            student_item_dict = config.get_student_item_dict()
            external_submissions = team_api.get_teammates_with_submissions_from_other_teams(
                config.course_id,
                student_item_dict["item_id"],
                team_info["team_id"],
                config.get_anonymous_user_ids_for_team(),
            )

            team_context["team_members_with_external_submissions"] = list_to_conversational_format(
                [config.get_username(submission["student_id"]) for submission in external_submissions]
            )
    except ObjectDoesNotExist:
        logger.error(
            "%s: User associated with anonymous_user_id %s can not be found.",
            str(config.location),
            config.get_student_item_dict()["student_id"],
        )
    except NoSuchServiceError:
        logger.error("%s: Teams service is unavailable", str(config.location))

    return team_context


def save_status(config, submission_info):
    """
    Return a string indicating whether the response has been saved.

    Returns:
        unicode
    """
    return config.translate("Draft saved!") if submission_info.has_saved else config.translate("Response not started.")


def get_submission_context(config, submission_info):
    """
    Determine the context needed when rendering the response (submission) step.

    Returns:
    * Context (dict) - Context used for rendering the submission
    """
    # Get ORA Metadata
    block_metadata = {
        "xblock_id": config.get_xblock_id(),
        "base_asset_url": config.base_asset_url,
    }

    # Get response config
    response_config = submission_info.response_config

    # Get user info / preferences
    user_preferences = get_user_preferences(config.user_service)
    user_config = {
        "has_real_user": config.has_real_user,
        "user_language": user_preferences["user_language"],
        "user_timezone": user_preferences["user_timezone"],
    }

    # Below here we determine submission status and template
    submission_context = {}
    team_submission_context = {}
    file_urls = submission_info.files.uploaded_files
    if file_urls:
        submission_context.update(file_urls)

    # Get access information (whether problem is closed) and reasons
    workflow_context = {}
    if submission_info.due_date:
        workflow_context["submission_due"] = submission_info.due_date

    # Response is unavailable (not yet open or past due date)
    if not submission_info.has_submitted and submission_info.problem_is_inaccessible:
        if submission_info.is_not_available_yet:
            workflow_context["submission_start"] = submission_info.start_date

    # Response not yet submitted: Get the saved response
    elif not submission_info.has_submitted:
        # Load the user/team saved response
        saved_response = submission_info.saved_response
        submission_context["saved_response"] = create_submission_dict(saved_response, config.prompts)

        if submission_info.is_team_assignment:
            team_context = get_team_submission_context(config)
            submission_context.update(team_context)

        # Determine UI states
        submission_context["save_status"] = save_status(config, submission_info)
        submission_context["enable_delete_files"] = True

    # Cancelled: Instructor has cancelled this response
    elif submission_info.has_been_cancelled:
        # Load user/team submission
        submission_context["student_submission"] = submission_info.student_submission

        # Get cancellation information
        workflow_context["workflow_cancellation"] = submission_info.cancellation_info

    # Done: Submitted and received final grade
    elif submission_info.has_received_final_grade:
        student_submission = submission_info.student_submission
        submission_context["student_submission"] = create_submission_dict(student_submission, config.prompts)

    # Submitted and waiting for a grade
    else:
        # Load user/team submission
        student_submission = submission_info.student_submission
        submission_context["student_submission"] = create_submission_dict(student_submission, config.prompts)

        # Get workflow context
        workflow_context["peer_incomplete"] = submission_info.peer_step_incomplete
        workflow_context["self_incomplete"] = submission_info.self_step_incomplete

    # Assemble context
    context = {
        **block_metadata,
        **response_config,
        **user_config,
        **workflow_context,
        **submission_context,
        **team_submission_context,
    }

    return context



"""Open Response available editors configuration"""

from django.conf import settings

from openassessment.xblock.load_static import LoadStatic


external_editors = getattr(settings, 'ORA_AVAILABLE_EDITORS', {})

default_editors = {
    'text': {
        'display_name': 'Simple Text Editor',
        'js': [LoadStatic.get_url('openassessment-editor-textarea.js')],
    },
    'tinymce': {
        'display_name': 'WYSIWYG Editor',
        'js': [LoadStatic.get_url('openassessment-editor-tinymce.js')],
    },
}

AVAILABLE_EDITORS = {}
AVAILABLE_EDITORS.update(external_editors)
AVAILABLE_EDITORS.update(default_editors)



"""
Student training step in the OpenAssessment XBlock.
"""
import logging

from openassessment.assessment.api import student_training
from openassessment.workflow.errors import AssessmentWorkflowError

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

messages = {
    "missing_selected": "Missing options_selected key in request.",
    "selected_must_be_dict": "options_selected must be a dictionary.",
    "could_not_check": "Your scores could not be checked.",
    "unexpected_error": "An unexpected error occurred.",
    "could_not_update": "Could not update workflow status.",
}


def training_assess(api_data, data):
    """
    Compare the scores given by the student with those given by the course author.
    If they match, update the training workflow.  The client can then reload this
    step to view the next essay or the completed step.

    Currently, we return a boolean indicating whether the student assessed correctly
    or not.  However, the student training API provides the exact criteria that the student
    scored incorrectly, as well as the "correct" options for those criteria.
    In the future, we may expose this in the UI to provide more detailed feedback.

    Args:
        data (dict): Must have the following keys:
            options_selected (dict): Dictionary mapping criterion names to option values.

    Returns:
        Dict with keys:
            * "success" (bool) indicating success or error
            * "msg" (unicode) containing additional information if an error occurs.
            * "correct" (bool) indicating whether the student scored the assessment correctly.

    """

    translate = api_data.config_data.translate
    submission_uuid = api_data.workflow_data.submission_uuid

    def failure_response(reason_key):
        return {"success": False, "msg": translate(messages[reason_key])}

    if "options_selected" not in data:
        return failure_response("missing_selected")
    if not isinstance(data["options_selected"], dict):
        return failure_response("selected_must_be_dict")

    # Check the student's scores against the course author's scores.
    # This implicitly updates the student training workflow (which example essay is shown)
    # as well as the assessment workflow (training/peer/self steps).
    try:
        corrections = student_training.assess_training_example(submission_uuid, data["options_selected"])
        api_data.config_data.publish_event(
            "openassessment.student_training_assess_example",
            {
                "submission_uuid": submission_uuid,
                "options_selected": data["options_selected"],
                "corrections": corrections,
            },
        )
    except student_training.StudentTrainingRequestError:
        msg = ("Could not check learner training scores for the learner with submission UUID {uuid}").format(
            uuid=submission_uuid
        )
        logger.warning(msg, exc_info=True)

        return failure_response("could_not_check")
    except student_training.StudentTrainingInternalError:
        return failure_response("could_not_check")
    except Exception:  # pylint: disable=broad-except
        return failure_response("unexpected_error")
    else:
        try:
            api_data.workflow_data.update_workflow_status()
        except AssessmentWorkflowError:
            logger.exception(translate(messages["could_not_update"]))
            return failure_response("could_not_update")
        return {
            "success": True,
            "msg": "",
            "corrections": corrections,
        }



""" Serializers for the staff_grader app """

from openassessment.staffgrader.serializers.submission_list import (
    MissingContextException, SubmissionListScoreSerializer, SubmissionListSerializer, TeamSubmissionListSerializer
)
from openassessment.staffgrader.serializers.submission_lock import SubmissionLockSerializer
from openassessment.staffgrader.serializers.assessments import (
    SubmissionDetailFileSerilaizer, AssessmentSerializer, AssessmentPartSerializer
)
